import logging

import astroplan
import click
from astropy import coordinates as coord

logger = logging.getLogger(__name__)


@click.command(
    epilog="""Check out the documentation at
               https://pyscope.readthedocs.io/en/latest/
               for more information."""
)
@click.option(
    "-s",
    "--source",
    type=str,
    default="",
    required=True,
    help="""Source name to resolve coordinates or a
              pair of ICRS ra,dec coordinates: hh:mm:ss,dd:mm:ss or any format
              that initializes an astropy.coordinates.SkyCoord object.""",
)
@click.option(
    "-f",
    "--filters",
    type=str,
    required=True,
    help="""Filters to use. Multiple filters can be
            specified by repeating the option, e.g. -f b -f g -f r""",
)
@click.option(
    "-e",
    "--exp-times",
    type=float,
    required=True,
    multiple=True,
    help="""Exposure times to use for each filter. Multiple exposure
            times can be specified by repeating the option, e.g. -e 10 -e 20.
            Must match length of filters.""",
)
@click.option(
    "-o",
    "--observer",
    nargs=2,
    type=str,
    default=("", ""),
    required=True,
    help="""Observer name and 3-character code.""",
)
@click.option(
    "-O",
    "--overlap",
    type=float,
    default=5,
    show_default=True,
    help="""Overlap between pointings (arcmin).""",
)
@click.option(
    "-F",
    "--fov",
    nargs=2,
    type=float,
    default=(20, 20),
    show_default=True,
    help="""Field of view (arcmin).""",
)
@click.option(
    "-g",
    "--grid-size",
    nargs=2,
    type=int,
    default=(3, 3),
    show_default=True,
    help="""Grid size (nrows, ncols).""",
)
@click.option(
    "-n",
    "--n-exp",
    type=int,
    default=1,
    show_default=True,
    help="""Number of exposures per filter and exposure time.""",
)
@click.option(
    "-r",
    "--readout",
    type=int,
    default=0,
    show_default=True,
    help="Readout mode as specified by the camera.",
)
@click.option(
    "-b", "--binning", type=str, default="1x1", show_default=True, help="Binning mode."
)
@click.option(
    "-u",
    "--ut-start",
    type=str,
    help="""UT start time of the observation in the ISOT format. If none
            given, the object will be scheduled for the best possible time.""",
)
@click.option(
    "-c",
    "--comment",
    type=str,
    default="",
    show_default=True,
    help="""Comment to be added to the schedule.""",
)
@click.option(
    "-w",
    "--write",
    type=str,
    default="",
    help="""Write a .sch file with the given name. If none given,
            a name is generated from the observing code.""",
)
@click.option(
    "-v",
    "--verbose",
    count=True,
    default=0,
    type=click.IntRange(0, 1),
    help="Increase verbosity",
)
@click.version_option()
def mk_mosaic_schedule_cli(
    source,
    filters,
    exp_times,
    observer,
    overlap=5,
    fov=(20, 20),
    grid_size=(3, 3),
    n_exp=1,
    readout=0,
    binning="1x1",
    ut_start=None,
    comment="",
    write=None,
    verbose=-1,
):
    """Generates an ObservingBlock set or a .sch file for mosaic imaging of wide fields.\b

    This script takes a source name to be resolved by `~astropy.coordinates.SkyCoord.from_name`
    or a pair of ICRS coordinates that initalizes a `~astropy.coordinates.SkyCoord` object and
    generates an `~astroplan.ObservingBlock` set or an .sch file.

    .. hint::

        Tools like `Telescopius <https://telescopius.com/telescope-simulator>`_ can be used to
        help you determine the calculate the `fov` and select the best `grid_size` and
        `overlap` for your particular telescope and target combination.

    Parameters
    ----------
    source : str
        Source name to resolve coordinates or a pair of ICRS ra,dec coordinates

    filters : str or tuple of str
        Filters to use. Multiple filters can be specified by a tuple, e.g. ('b', 'g', 'r')

    exp_times : float or tuple of float
        Exposure times to use. Multiple exposure times can be specified by a tuple, e.g. (10, 20)

    observer : tuple of str
        Observer name and 3-character code, e.g. ('Walter Golay', 'XWG')

    overlap : float, default=5
        Overlap between pointings (arcmin).

    fov : tuple of float, default=(20, 20)
        Field of view (arcmin).

    grid_size : tuple of int, default=(3, 3)
        Grid size (nrows, ncols).

    n_exp : int, default=1
        Number of exposures per filter and exposure time.

    readout : int, default=0
        Readout mode as specified by the camera.

    binning : str, default='1x1'
        Binning mode.

    ut_start : str, optional
        UT start time of the observation in the ISOT format. If none given,
        the object will be scheduled for the best possible time.

    comment : str, optional
        Comment to be added to the schedule.

    write : str, optional
        Write a .sch file with the given name. If none given from the command line,
        a name is generated from the observing code. Otherwise, a file will not
        be written and the ObservingBlock set will be returned.

    verbose : int, {-1, 0, 1}, default=-1
        Verbosity level. 0: no debug messages, 1: debug messages. Default set to -1
        for API use and 0 for CLI use.

    Returns
    -------
    obsblocks : `~astroplan.ObservingBlock` tuple or None
        Set of ObservingBlocks for the mosaic observation.

    Raises
    ------
    `click.BadParameter`

    See Also
    --------
    pyscope.telrun.schedtel
    astropy.coordinates.SkyCoord

    Examples
    --------
    API use:

    .. doctest::

        >>> from pyscope.telrun import mk_mosaic_schedule, schedtel
        >>> blocks = mk_mosaic_schedule('M31', 'g', 30, ('Walter Golay', 'XWG'))

    """

    if len(filters) != len(exp_times):
        raise click.BadParameter(
            "Number of filters must match number of exposure times"
        )

    if verbose > -1:
        logger.setLevel(int(10 * (2 - verbose)))
        logger.addHandler(logging.StreamHandler())
    logger.debug(f"Verbosity level set to {verbose}")
    logger.debug(
        f"""mk_mosaic_schedule_cli(source={source}, filters={filters},
        exp_times={exp_times}, observer={observer}, overlap={overlap}, fov={fov},
        grid_size={grid_size}, n_exp={n_exp}, readout={readout}, binning={binning},
        ut_start={ut_start}, comment={comment}, write={write}, verbose={verbose})"""
    )

    if len(source.split(",")) == 2:
        ra, dec = source.split(",")
        center = coord.SkyCoord(ra, dec, unit=("hourangle", "deg"))
    else:
        center = coord.SkyCoord.from_name(source)
    logger.debug(
        f"Resolved coordinates: {center.to_string('hmsdms', sep=':', precision=1)}"
    )

    # Generate grid
    dra = (fov[0] - overlap) / 60 / np.cos(center.dec.deg * np.pi / 180.0)
    ddec = (fov[1] - overlap) / 60
    ra_list = np.linspace(
        center.ra.deg - (grid_size[1] - 1) * dra / 2.0,
        center.ra.deg + (grid_size[1] - 1) * dra / 2.0,
        grid_size[1],
        endpoint=True,
    )
    dec_list = np.linspace(
        center.dec.deg - (grid_size[0] - 1) * ddec / 2.0,
        center.dec.deg + (grid_size[0] - 1) * ddec / 2.0,
        grid_size[0],
        endpoint=True,
    )
    grid_list = np.array(np.meshgrid(ra_list, dec_list)).T.reshape(-1, 2)

    # Generate ObservingBlocks
    obsblocks = []
    for i, pos in enumerate(grid_list):
        for j in range(len(filters)):
            filt = filters[j]
            exp = exposures[j]
            constraints = None

            if utstart is not None:
                constraints = [
                    [
                        astroplan.constraints.TimeConstraint(
                            utstart + (j + k * len(j)) * exp * u.second - schederr,
                            utstart + (j + k * len(j)) * exp * u.second + schederr,
                        )
                    ]
                    for k in range(n_exp)
                ]

            for j in range(n_exp):
                obsblocks.append(
                    astroplan.ObservingBlock(
                        astroplan.FixedTarget(
                            coord.SkyCoord(pos[0], pos[1], unit="deg")
                        ),
                        exp_times[j],
                        priority=1,
                        configuration={
                            "observer": observer[0],
                            "code": observer[1],
                            "title": f"Mosaic imaging of {source}",
                            "filter": filters[j],
                            "exposure": exp_times[j],
                            "n_exp": 1,
                            "readout": readout,
                            "binning": binning.split("x"),
                            "comment": comment,
                        },
                        constraints=constraints[j] if constraints is not None else None,
                        name=f"{source}_{i}_{j}",
                    )
                )
                logger.debug(
                    f"""Created ObservingBlock: {blocks[-1].target},
                    {blocks[-1].duration}, {blocks[-1].priority},
                    {blocks[-1].name}, {blocks[-1].constraints},
                    {blocks[-1].configuration}"""
                )

    if write is not None:
        if write == "":
            write = f"{observer[1]}{strftime('%j', localtime())}.sch"
        with open(write, "w") as f:
            f.write(f"tit 'Mosaic imaging of {source}'\n")
            f.write(f"obs '{observer[0]}'\n")
            f.write("# Schedule generated by mk-mosaic-sched'\n")
            f.write(f"filt {','.join(filters)} exp {','.join(exp_times)}\n\n")
            for block in obsblocks:
                ra_str, dec_str = block.target.to_string(
                    style="hmsdms",
                    precision=3,
                    sep=":",
                    decimal=True,
                ).split()
                f.write(
                    f"""sou {block.name} ra {ra_str} dec {dec_str} repe {n_exp} uts {utstart} rea {readout} bin {binning} com {comment}\n"""
                )

    if verbose > -1:
        logger.info("============================================================")
        logger.info("Mosaic of %s" % source)
        logger.info(
            "%i filters, %.0f minutes per filter sequence"
            % (len(filters), sum(exp_times))
        )
        logger.info("Repeat each filter sequence %i times" % n_exp)
        logger.info(
            "%i x %i field, %i fields total"
            % (grid_size[0], grid_size[1], grid_size[0] * grid_size[1])
        )
        logger.info(
            "%i images, total observing time %0.1f hours"
            % (len(obsblocks), len(filters) * sum(exp_times) * n_exp / 3600.0)
        )
        logger.info("Output file: %s" % write)
        logger.info("============================================================")


mk_mosaic_schedule = mk_mosaic_schedule_cli.callback



import logging
import urllib.request

from ..utils import _get_number_from_line
from .observing_conditions import ObservingConditions

logger = logging.getLogger(__name__)


class HTMLObservingConditions(ObservingConditions):
    def __init__(
        self,
        url,
        cloud_cover_keyword=b"CLOUDCOVER",
        cloud_cover_units=b"%",
        cloud_cover_numeric=True,
        dew_point_keyword=b"DEWPOINT",
        dew_point_units=b"F",
        dew_point_numeric=True,
        humidity_keyword=b"HUMIDITY",
        humidity_units=b"%",
        humidity_numeric=True,
        pressure_keyword=b"PRESSURE",
        pressure_units=b"inHg",
        pressure_numeric=True,
        rain_rate_keyword=b"RAINRATE",
        rain_rate_units=b"inhr",
        rain_rate_numeric=True,
        sky_brightness_keyword=b"SKYBRIGHTNESS",
        sky_brightness_units="magdeg2",
        sky_brightness_numeric=True,
        sky_quality_keyword=b"SKYQUALITY",
        sky_quality_units="",
        sky_quality_numeric=True,
        sky_temperature_keyword=b"SKYTEMPERATURE",
        sky_temperature_units=b"F",
        sky_temperature_numeric=True,
        star_fwhm_keyword=b"STARFWHM",
        star_fwhm_units=b"arcsec",
        star_fwhm_numeric=True,
        temperature_keyword=b"TEMPERATURE",
        temperature_units=b"F",
        temperature_numeric=True,
        wind_direction_keyword=b"WINDDIRECTION",
        wind_direction_units=b"EofN",
        wind_direction_numeric=True,
        wind_gust_keyword=b"WINDGUST",
        wind_gust_units="mph",
        wind_gust_numeric=True,
        wind_speed_keyword=b"WINDSPEED",
        wind_speed_units="mph",
        wind_speed_numeric=True,
        last_updated_keyword=b"LASTUPDATED",
        last_updated_units="",
        last_updated_numeric=True,
    ):
        logger.debug(
            f"""HTMLObservingConditions.__init__(
                    {url},
                    cloud_cover_keyword={cloud_cover_keyword},
                    cloud_cover_units={cloud_cover_units},
                    cloud_cover_numeric={cloud_cover_numeric},
                    dew_point_keyword={dew_point_keyword},
                    dew_point_units={dew_point_units},
                    dew_point_numeric={dew_point_numeric},
                    humidity_keyword={humidity_keyword},
                    humidity_units={humidity_units},
                    humidity_numeric={humidity_numeric},
                    pressure_keyword={pressure_keyword},
                    pressure_units={pressure_units},
                    pressure_numeric={pressure_numeric},
                    rain_rate_keyword={rain_rate_keyword},
                    sky_brightness_keyword={sky_brightness_keyword},
                    sky_brightness_units={sky_brightness_units},
                    sky_brightness_numeric={sky_brightness_numeric},
                    sky_quality_keyword={sky_quality_keyword},
                    sky_quality_units={sky_quality_units},
                    sky_quality_numeric={sky_quality_numeric},
                    sky_temperature_keyword={sky_temperature_keyword},
                    sky_temperature_units={sky_temperature_units},
                    sky_temperature_numeric={sky_temperature_numeric},
                    star_fwhm_keyword={star_fwhm_keyword},
                    star_fwhm_units={star_fwhm_units},
                    star_fwhm_numeric={star_fwhm_numeric},
                    temperature_keyword={temperature_keyword},
                    temperature_units={temperature_units},
                    temperature_numeric={temperature_numeric},
                    wind_direction_keyword={wind_direction_keyword},
                    wind_direction_units={wind_direction_units},
                    wind_direction_numeric={wind_direction_numeric},
                    wind_gust_keyword={wind_gust_keyword},
                    wind_gust_units={wind_gust_units},
                    wind_gust_numeric={wind_gust_numeric},
                    wind_speed_keyword={wind_speed_keyword},
                    wind_speed_units={wind_speed_units},
                    wind_speed_numeric={wind_speed_numeric},
                    last_updated_keyword={last_updated_keyword},
                    last_updated_units={last_updated_units},
                    last_updated_numeric={last_updated_numeric}) called"""
        )

        self._url = url

        self._cloud_cover = None
        self._cloud_cover_keyword = cloud_cover_keyword
        self._cloud_cover_units = cloud_cover_units
        self._cloud_cover_numeric = cloud_cover_numeric
        self._dew_point = None
        self._dew_point_keyword = dew_point_keyword
        self._dew_point_units = dew_point_units
        self._dew_point_numeric = dew_point_numeric
        self._humidity = None
        self._humidity_keyword = humidity_keyword
        self._humidity_units = humidity_units
        self._humidity_numeric = humidity_numeric
        self._pressure = None
        self._pressure_keyword = pressure_keyword
        self._pressure_units = pressure_units
        self._pressure_numeric = pressure_numeric
        self._rain_rate = None
        self._rain_rate_keyword = rain_rate_keyword
        self._rain_rate_units = rain_rate_units
        self._rain_rate_numeric = rain_rate_numeric
        self._sky_brightness = None
        self._sky_brightness_keyword = sky_brightness_keyword
        self._sky_brightness_units = sky_brightness_units
        self._sky_brightness_numeric = sky_brightness_numeric
        self._sky_quality = None
        self._sky_quality_keyword = sky_quality_keyword
        self._sky_quality_units = sky_quality_units
        self._sky_quality_numeric = sky_quality_numeric
        self._sky_temperature = None
        self._sky_temperature_keyword = sky_temperature_keyword
        self._sky_temperature_units = sky_temperature_units
        self._sky_temperature_numeric = sky_temperature_numeric
        self._star_fwhm = None
        self._star_fwhm_keyword = star_fwhm_keyword
        self._star_fwhm_units = star_fwhm_units
        self._star_fwhm_numeric = star_fwhm_numeric
        self._temperature = None
        self._temperature_keyword = temperature_keyword
        self._temperature_units = temperature_units
        self._temperature_numeric = temperature_numeric
        self._wind_direction = None
        self._wind_direction_keyword = wind_direction_keyword
        self._wind_direction_units = wind_direction_units
        self._wind_direction_numeric = wind_direction_numeric
        self._wind_gust = None
        self._wind_gust_keyword = wind_gust_keyword
        self._wind_gust_units = wind_gust_units
        self._wind_gust_numeric = wind_gust_numeric
        self._wind_speed = None
        self._wind_speed_keyword = wind_speed_keyword
        self._wind_speed_units = wind_speed_units
        self._wind_speed_numeric = wind_speed_numeric
        self._last_updated = None
        self._last_updated_keyword = last_updated_keyword
        self._last_updated_units = last_updated_units
        self._last_updated_numeric = last_updated_numeric

        self.Refresh()

    def Refresh(self):
        logger.debug(f"HTMLObservingConditions.Refresh() called")
        stream = urllib.request.urlopen(self._url)
        lines = stream.readlines()

        for line in lines:
            cloud_cover = _get_number_from_line(
                line,
                self._cloud_cover_keyword,
                self._cloud_cover_units,
                self._cloud_cover_numeric,
            )
            dew_point = _get_number_from_line(
                line,
                self._dew_point_keyword,
                self._dew_point_units,
                self._dew_point_numeric,
            )
            humidity = _get_number_from_line(
                line,
                self._humidity_keyword,
                self._humidity_units,
                self._humidity_numeric,
            )
            pressure = _get_number_from_line(
                line,
                self._pressure_keyword,
                self._pressure_units,
                self._pressure_numeric,
            )
            rain_rate = _get_number_from_line(
                line,
                self._rain_rate_keyword,
                self._rain_rate_units,
                self._rain_rate_numeric,
            )
            sky_brightness = _get_number_from_line(
                line,
                self._sky_brightness_keyword,
                self._sky_brightness_units,
                self._sky_brightness_numeric,
            )
            sky_quality = _get_number_from_line(
                line,
                self._sky_quality_keyword,
                self._sky_quality_units,
                self._sky_quality_numeric,
            )
            sky_temperature = _get_number_from_line(
                line,
                self._sky_temperature_keyword,
                self._sky_temperature_units,
                self._sky_temperature_numeric,
            )
            star_fwhm = _get_number_from_line(
                line,
                self._star_fwhm_keyword,
                self._star_fwhm_units,
                self._star_fwhm_numeric,
            )
            temperature = _get_number_from_line(
                line,
                self._temperature_keyword,
                self._temperature_units,
                self._temperature_numeric,
            )
            wind_direction = _get_number_from_line(
                line,
                self._wind_direction_keyword,
                self._wind_direction_units,
                self._wind_direction_numeric,
            )
            wind_gust = _get_number_from_line(
                line,
                self._wind_gust_keyword,
                self._wind_gust_units,
                self._wind_gust_numeric,
            )
            wind_speed = _get_number_from_line(
                line,
                self._wind_speed_keyword,
                self._wind_speed_units,
                self._wind_speed_numeric,
            )
            last_updated = _get_number_from_line(
                line,
                self._last_updated_keyword,
                self._last_updated_units,
                self._last_updated_numeric,
            )

            if cloud_cover is not None:
                self._cloud_cover = cloud_cover
            if dew_point is not None:
                self._dew_point = dew_point
            if humidity is not None:
                self._humidity = humidity
            if pressure is not None:
                self._pressure = pressure
            if rain_rate is not None:
                self._rain_rate = rain_rate
            if sky_brightness is not None:
                self._sky_brightness = sky_brightness
            if sky_quality is not None:
                self._sky_quality = sky_quality
            if sky_temperature is not None:
                self._sky_temperature = sky_temperature
            if star_fwhm is not None:
                self._star_fwhm = star_fwhm
            if temperature is not None:
                self._temperature = temperature
            if wind_direction is not None:
                self._wind_direction = wind_direction
            if wind_gust is not None:
                self._wind_gust = wind_gust
            if wind_speed is not None:
                self._wind_speed = wind_speed
            if last_updated is not None:
                self._last_updated = last_updated

    def SensorDescription(self, PropertyName):
        logger.debug("HTMLObservingConditions.SensorDescription({PropertyName}) called")
        return eval(f"self._{PropertyName.lower()}_keyword")

    def TimeSinceLastUpdate(self, PropertyName):
        logger.debug(
            "HTMLObservingConditions.TimeSinceLastUpdate({PropertyName}) called"
        )
        stream = urllib.request.urlopen(self._url)
        lines = stream.readlines()

        for line in lines:
            last_updated = _get_number_from_line(
                line,
                self._last_updated_keyword,
                self._last_updated_units,
                self._last_updated_numeric,
            )
            if last_updated is not None:
                self._last_updated = last_updated

        return self.LastUpdated

    @property
    def AveragePeriod(self):
        logger.debug("HTMLObservingConditions.AveragePeriod property called")
        return

    @AveragePeriod.setter
    def AveragePeriod(self, value):
        logger.debug(f"HTMLObservingConditions.AveragePeriod({value}) called")
        return

    @property
    def CloudCover(self):
        logger.debug("HTMLObservingConditions.CloudCover property called")
        return self._cloud_cover

    @property
    def DewPoint(self):
        logger.debug("HTMLObservingConditions.DewPoint property called")
        return self._dew_point

    @property
    def Humidity(self):
        logger.debug("HTMLObservingConditions.Humidity property called")
        return self._humidity

    @property
    def Pressure(self):
        logger.debug("HTMLObservingConditions.Pressure property called")
        return self._pressure

    @property
    def RainRate(self):
        logger.debug("HTMLObservingConditions.RainRate property called")
        return self._rain_rate

    @property
    def SkyBrightness(self):
        logger.debug("HTMLObservingConditions.SkyBrightness property called")
        return self._sky_brightness

    @property
    def SkyQuality(self):
        logger.debug("HTMLObservingConditions.SkyQuality property called")
        return self._sky_quality

    @property
    def SkyTemperature(self):
        logger.debug("HTMLObservingConditions.SkyTemperature property called")
        return self._sky_temperature

    @property
    def StarFWHM(self):
        logger.debug("HTMLObservingConditions.StarFWHM property called")
        return self._star_fwhm

    @property
    def Temperature(self):
        logger.debug("HTMLObservingConditions.Temperature property called")
        return self._temperature

    @property
    def WindDirection(self):
        logger.debug("HTMLObservingConditions.WindDirection property called")
        return self._wind_direction

    @property
    def WindGust(self):
        logger.debug("HTMLObservingConditions.WindGust property called")
        return self._wind_gust

    @property
    def WindSpeed(self):
        logger.debug("HTMLObservingConditions.WindSpeed property called")
        return self._wind_speed

    @property
    def LastUpdated(self):
        logger.debug("HTMLObservingConditions.LastUpdated property called")
        return self._last_updated



#!/usr/bin/env python

# sexphot2
# Computes 2-color [G,R] photometric magnitudes using sextractor, plots light curves for G, R filters and g-r color index
# Optionally checks magnitudes using SDSS

# N.B. Requires sextractor!
# [command line sex, config file location defaults to /usr/local/sextractor/default.sex]
# v. 1.0 11 Mar 2017
# v. 1.1 26 Mar 2017  - add fwhm_range in opts
# v. 1.2 31 May 2017  - check photometry
# v. 1.3 13 Jun 2017  - fix problem with some epochs having nan magnitudes

vers = "1.3 (13 Jun 2017)"

import glob
import itertools
import os
import re
import sys
import warnings
from optparse import OptionParser

import matplotlib.pyplot as plt
import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.io.fits import getheader, setval, update
from astropy.time import Time
from astroquery.sdss import SDSS
from matplotlib.pyplot import cm
from scipy.optimize import minimize

# Avoid annoying warning about matplotlib building the font cache
warnings.filterwarnings("ignore")

# Sextractor config file path
sex_path = "/usr/local/sextractor/default.sex"


def get_args():
    global parser
    parser = OptionParser(
        description="Program %prog plots 2-color (Sloan g,r) absolute photometric magnitudes and color index g-r using sextractor, ZP mag",
        version="%s" % vers,
    )
    parser.add_option(
        "-s",
        dest="sigma",
        metavar="sigma",
        action="store",
        type=float,
        default=5,
        help="Sextractor detection threshold [default 5]",
    )
    parser.add_option(
        "-c",
        dest="config",
        metavar="config",
        action="store",
        help="phot config file name [no default]",
    )
    parser.add_option(
        "-d",
        dest="datafile",
        metavar="outfile",
        action="store",
        default="",
        help="Output csv file name",
    )
    parser.add_option(
        "-f",
        dest="fwhm_range",
        metavar="fwhm_range",
        action="store",
        default="1.4,5",
        help="FWHM max (pixels)   [default 1.4,5]",
    )
    parser.add_option(
        "-l",
        dest="line",
        metavar="line",
        action="store_true",
        default=False,
        help="Plot median line",
    )
    parser.add_option(
        "-p",
        dest="plot",
        metavar="plot",
        action="store_true",
        default=True,
        help="Plot solution",
    )
    parser.add_option(
        "-J",
        dest="jdrange",
        metavar="jdrange",
        action="store",
        default="0,0",
        help="JD range (JDmin, JDmax)",
    )
    parser.add_option(
        "-v",
        dest="verbose",
        metavar="Verbose",
        action="store_true",
        default=False,
        help="Verbose output",
    )
    parser.add_option(
        "-y",
        dest="ywidth",
        metavar="ywidth",
        action="store",
        type=float,
        default=2.0,
        help="Differential plot width [mags, default 2 mag]",
    )
    return parser.parse_args()


def get_hdrdata(ftsfile):
    hdr = getheader(ftsfile, 0)
    jd = hdr["JD"]
    date = hdr["DATE-OBS"]
    exptime = hdr["EXPTIME"]
    filter = hdr["FILTER"][0]
    airmass = hdr["AIRMASS"]
    if "ZMAG" in hdr:
        zp = hdr["ZMAG"]
        zperr = hdr["ZMAGERR"]
    else:
        zp = 0
        zperr = 0
    nbin = hdr["XBINNING"]  # Assume same for y binning
    arcsec_pixel = np.abs(hdr["CDELT1"] * 3600.0)
    return jd, date, exptime, filter, arcsec_pixel, airmass, nbin, zp, zperr


def get_sexinfo(sexname, exptime, scale):
    global fwhm_min, fwhm_max
    fn = open(sexname, "r")
    lines = fn.readlines()[15:]
    Nr = []
    Ra = []
    Dec = []
    Snr = []
    Flux = []
    Fluxerr = []
    Fwhm = []
    V = []
    Verr = []
    for line in lines:
        (
            nr,
            dum,
            dum,
            flux,
            fluxerr,
            x_pix,
            y_pix,
            ra_deg,
            dec_deg,
            profile_x,
            profile_y,
            pa,
            fwhm_pixel,
            dum,
            flag,
        ) = [float(x) for x in line.split()]
        v = -2.5 * np.log10(flux / exptime)
        if fluxerr == 0:
            continue
        snr = flux / fluxerr
        verr = 2.5 * (fluxerr / flux)  # Expanding log10(1+x) ~ 2.5x
        Nr.append(nr)
        Ra.append(ra_deg)
        Dec.append(dec_deg)
        Flux.append(flux)
        Fluxerr.append(fluxerr)
        Fwhm.append(fwhm_pixel * np.abs(scale))
        Snr.append(snr)
        V.append(v)
        Verr.append(verr)
    fn.close()
    n1 = len(V)
    # Trim list to stars by restricting fwhm values
    A = zip(Ra, Dec, Snr, Flux, Fluxerr, Fwhm, V, Verr)
    B = []
    for j in range(len(A)):
        if fwhm_min < A[j][5] < fwhm_max:
            B.append(A[j])
    Ra, Dec, Snr, Flux, Fluxerr, Fwhm, V, Verr = zip(*B)
    n2 = len(V)
    if verbose:
        print(
            "Warning, skipped %i stars because they exceeded allowed FWHM range (%.1f,%.1f), use option -f to change"
            % (n1 - n2, fwhm_min, fwhm_max)
        )
    V = np.array(V)
    Verr = np.array(Verr)
    return Nr, Ra, Dec, Snr, Flux, Fluxerr, Fwhm, V, Verr


def get_sdss_magnitudes(ra, dec):
    # Query SDSS online photometric catalog for u,g,r,i,z magnitudes; ra,deg in degrees (ICRS, 2000)
    pos = SkyCoord(ra, dec, unit=(u.deg, u.deg), frame="icrs")
    ids = SDSS.query_region(
        pos, radius=5 * u.arcsec, fields=["ra", "dec", "clean", "u", "g", "r", "i", "z"]
    )  # defaults to 2 arcsec search
    u1 = g = r = i = z = np.nan
    if ids != None:
        for id in ids:
            if (
                id["clean"] == 1 and id["g"] < 20.0
            ):  # Only accept photometry with clean flags & reject very faint stars
                u1 = id["u"]
                g = id["g"]
                r = id["r"]
                i = id["i"]
                z = id["z"]
                return u1, g, r, i, z
    return u1, g, r, i, z


def trim(indices, A):
    # Trims arrays packed in A, dropping elements with given indices
    B = []
    for a in A:
        B.append(np.delete(a, indices))
    return B


def get_magnitudes(Ra, Dec, Nr_sex, Ra_sex, Dec_sex, max_diff, Mag_sex, Mag_sex_err):
    N1 = len(Ra)
    N2 = len(Ra_sex)
    Mag = np.empty(N1) * np.nan
    Mag_err = np.empty(N1) * np.nan
    Nr = np.empty(N1) * np.nan
    for j in range(N1):
        for k in range(N2):
            dra = np.abs(Ra[j] - Ra_sex[k])
            ddec = np.abs(Dec[j] - Dec_sex[k])
            if dra < max_diff and ddec < max_diff:
                Mag[j] = Mag_sex[k]
                Mag_err[j] = Mag_sex_err[k]
                Nr[j] = Nr_sex[k]
    Mag = np.array(Mag)
    Mag_err = np.array(Mag_err)
    return Nr, Mag, Mag_err


def parse_config(config_file):
    Objects = []
    Filters = []
    Ftsfiles_G = []
    Ftsfiles_R = []
    Ra_hms = []
    Dec_dms = []
    Ra_deg = []
    Dec_deg = []
    if not os.path.isfile(config_file):
        sys.exit("Configuration file %s does not exist, try again" % config_file)
    else:
        fn = open(config_file, "r")
        lines = fn.readlines()
        fn.close()
        for line in lines:
            line = line.split()
            if line == []:
                continue  # Skip blank lines
            elif line[0] == "G":
                Ftsfiles_G.append(line[1])
            elif line[0] == "R":
                Ftsfiles_R.append(line[1])
            elif line[0] == "S":
                object, ra_hms, dec_dms = line[1:4]
                Objects.append(object)
                Ra_hms.append(ra_hms)
                Dec_dms.append(dec_dms)
                c = SkyCoord(ra_hms, dec_dms, unit=(u.hourangle, u.deg), frame="icrs")
                Ra_deg.append(c.ra.deg)
                Dec_deg.append(c.dec.deg)
            elif line[0] == "T":
                title = " ".join(line[1:])
    return Objects, Ftsfiles_G, Ftsfiles_R, Ra_hms, Dec_dms, Ra_deg, Dec_deg, title


def get_star_info(Ftsfiles, Filter, Cal_vals):
    JD = []
    Date = []
    Mag_all = []
    Mag_err_all = []
    FitsFile_all = []
    Nr_all = []

    for ftsfile in Ftsfiles:
        # Get useful header info [NB not currently using nbin]
        jd, date, exptime, filter, scale, airmass, nbin, zp, zperr = get_hdrdata(
            ftsfile
        )

        # If wrong filter, skip
        if filter != Filter:
            if verbose:
                print(
                    "%s: Wrong filter [expecting %s, got %s], skipping"
                    % (ftsfile, Filter, filter)
                )
            continue

        # Run sextractor
        if verbose:
            print(
                "Running sextractor on %s with detection threshold = %.1f sigma"
                % (ftsfile, detect_threshold)
            )
        sname = "%s_%s.sexout" % (os.path.basename(ftsfile).split(".")[0], Filter)
        os.system(
            "sex %s -c %s -CATALOG_NAME %s -DETECT_THRESH %.1f -VERBOSE_TYPE QUIET"
            % (ftsfile, sex_path, sname, detect_threshold)
        )

        # Get position, magnitude info for each listed star in output file
        (
            Nr_sex,
            Ra_sex,
            Dec_sex,
            Snr,
            Flux,
            Fluxerr,
            Fwhm_sex,
            Mag_sex,
            Mag_sex_err,
        ) = get_sexinfo(sname, exptime, scale)
        nobs = len(Ra_sex)
        if verbose:
            print("Sextractor found %i stars" % nobs)

        # Get magnitudes for target objects using position match to sextractor output
        Nr, Mag, Mag_err = get_magnitudes(
            Ra_deg, Dec_deg, Nr_sex, Ra_sex, Dec_sex, max_diff, Mag_sex, Mag_sex_err
        )

        # Convert to magnitude by adding ZP. Use user-supplied ZP if specified
        if zp > 0:
            ZP = zp
            k = Cal_vals[Filter][1]
        else:
            ZP, k = Cal_vals[Filter]
            if verbose:
                print(
                    "WARNING: Using default zero-point for %s filter: (ZP = %.2f)"
                    % (filter, ZP)
                )
        # correct for airmass, assume average color correction 0.1
        Mag -= k * airmass - 0.1
        Mag += ZP
        # Add to array, but only if all stars detected
        # if not np.isnan(Mag).any():
        if True:
            JD.append(jd)
            Date.append(date)
            Mag_all.append(Mag)
            Mag_err_all.append(Mag_err)
            FitsFile_all.append(ftsfile)
            Nr_all.append(Nr)
    # Sort by JD
    JD, FitsFile_all, Date, Nr_all, Mag_all, Mag_err_all = (
        list(x)
        for x in zip(*sorted(zip(JD, FitsFile_all, Date, Nr_all, Mag_all, Mag_err_all)))
    )

    # Convert to numpy arrays
    Mag = np.array(Mag_all)
    Mag_err = np.array(Mag_err_all)

    # Calculate median differential magnitudes
    Medians = np.nanmedian(Mag, axis=0)
    Stds = np.nanstd(Mag, axis=0)
    # print 'Medians = %s' % ( (' '.join('%5.2f' % x for x in Medians)))
    return JD, Date, FitsFile_all, Mag, Mag_err, Medians, Stds


def mk_plot(
    j,
    JD0,
    Date0,
    object,
    Ra_hms,
    Dec_dms,
    mjd_g,
    mjd_r,
    G_mag,
    G_mag_err,
    R_mag,
    R_mag_err,
    ywidth,
    line,
):
    fig = plt.figure(j, figsize=(12, 8))
    ax = fig.add_subplot(111)
    ym = np.nanmean(np.append(G_mag, R_mag))
    ymin = ym + ywidth / 2.0
    ymax = ymin - ywidth
    plt.errorbar(mjd_g, G_mag, yerr=G_mag_err, fmt="bs", label="Sloan g")
    plt.errorbar(mjd_r, R_mag, yerr=R_mag_err, fmt="rs", label="Sloan r")
    plt.suptitle(title, fontsize=14)
    plt.title("%s [RA: %s, Dec: %s]" % (object, Ra_hms, Dec_dms), fontsize=12)
    if jdmin != 0.0:
        plt.xlim(jdmin - JD0, jdmax - JD0)
    plt.legend(loc=2)
    plt.ylim(ymin, ymax)
    if line:
        xmin, xmax = ax.get_xlim()
        xline = [xmin, xmax]
        ym_g = np.nanmedian(G_mag)
        ym_r = np.nanmedian(R_mag)
        std_g = np.nanstd(G_mag)
        std_r = np.nanstd(R_mag)
        yline_g = [ym_g, ym_g]
        yline_r = [ym_r, ym_r]
        plt.plot(xline, yline_g, linestyle="dashed", color="blue")
        plt.plot(xline, yline_r, linestyle="dashed", color="red")
        txt = "Median g = %.2f +/- %.2f,  r = %.2f +/- %.2f, g-r = %.2f" % (
            ym_g,
            std_g,
            ym_r,
            std_r,
            ym_g - ym_r,
        )
        plt.text(
            0.05,
            0.05,
            txt,
            fontsize=12,
            transform=ax.transAxes,
            bbox=dict(facecolor="white", alpha=0.5),
        )

    UTDate, UTTime = Date0.split("T")
    plt.xlabel("Days since JD %.5f (%s, %s UT)" % (JD0, UTDate, UTTime))
    plt.ylabel("Magnitude at airmass = 0")
    plt.grid(True)
    plot_title = "%s_lc.pdf" % object
    plt.savefig(plot_title)
    print("Saved differential l.c. plot %s" % plot_title)
    plt.close(fig)


# ======== MAIN ================

# Max difference: config vs Sex position [deg]
max_diff = 5 / 3600.0

# Define dictionary of zero-point values and extinction for filters [guesses except for G, R]
Cal_vals = {
    "N": (21.5, 0.20),
    "B": (21.0, 0.35),
    "G": (22.2, 0.28),
    "V": (20.5, 0.20),
    "R": (21.8, 0.12),
    "W": (20.66, 0.05),
}

# Get command  line arguments, assign parameter values
(opts, args) = get_args()

if not opts.config:
    parser.error("config file (-c) not given, try again")
config_file = opts.config

detect_threshold = opts.sigma  # Sextractor detection threshold [sigma]
plot = opts.plot  # Plot various things
csvfile = opts.datafile  # optional CSV output filename
jdmin, jdmax = [float(x) for x in opts.jdrange.split(",")]  # Julian date range
line = opts.line  # Plot median line
verbose = opts.verbose  # Print diagnostics, more
ywidth = opts.ywidth  # Differential plot width, magnitudes
fwhm_min, fwhm_max = [
    float(x) for x in opts.fwhm_range.split(",")
]  # Maximum allowed FWHM (pixels)

# Parse configuration file
Objects, Ftsfiles_G, Ftsfiles_R, Ra_hms, Dec_dms, Ra_deg, Dec_deg, title = parse_config(
    config_file
)
nstar = len(Objects)


# Expand filenames if needed
if "*" in Ftsfiles_G[0] or "?" in Ftsfiles_G[0]:
    Ftsfiles_G = glob.glob(Ftsfiles_G[0])
if "*" in Ftsfiles_R[0] or "?" in Ftsfiles_R[0]:
    Ftsfiles_R = glob.glob(Ftsfiles_R[0])


Filters = ["G", "R"]
for Filter in Filters:
    if Filter == "G":
        (
            JD_G,
            Date_G,
            FitsFiles_all_G,
            Mag_G,
            Mag_err_G,
            Medians_G,
            Stds_G,
        ) = get_star_info(Ftsfiles_G, Filter, Cal_vals)
    elif Filter == "R":
        (
            JD_R,
            Date_R,
            FitsFiles_all_R,
            Mag_R,
            Mag_err_R,
            Medians_R,
            Stds_R,
        ) = get_star_info(Ftsfiles_R, Filter, Cal_vals)
"""
print 'G'
for k in range(len(JD_G)):
	print JD_G[k], Mag_G[k]
print 'R'
for k in range(len(JD_R)):
	print JD_R[k], Mag_R[k]
"""

# Plot light curves

JD0 = min(JD_G[0], JD_R[0])
if JD0 == JD_G[0]:
    Date0 = Date_G[0]
else:
    Date0 = Date_R[0]
mjd_g = [x - JD0 for x in JD_G]
mjd_r = [x - JD0 for x in JD_R]
marker = itertools.cycle(["s", "d", "p", "*", "8"])

# First plot all stars on same plot
plt.figure(1, figsize=(12, 8))

params = {"legend.fontsize": 10}
plt.rcParams.update(params)

G_mag = np.transpose(Mag_G)
G_mag_err = np.transpose(Mag_err_G)
R_mag = np.transpose(Mag_R)
R_mag_err = np.transpose(Mag_err_R)

# Correct for color

for k in range(nstar):
    mark = marker.next()
    plt.errorbar(
        mjd_g,
        G_mag[k],
        yerr=G_mag_err[k],
        marker=mark,
        markersize=8,
        c="b",
        label="%s [G]" % Objects[k],
    )
    plt.errorbar(
        mjd_r,
        R_mag[k],
        yerr=R_mag_err[k],
        marker=mark,
        markersize=8,
        c="r",
        label="%s [R]" % Objects[k],
    )

plt.title(title)
plt.legend(loc=2)
plt.ylim(18, 12)
if jdmin != 0.0:
    plt.xlim(jdmin - JD[0], jdmax - JD[0])
plt.ylabel("%s Magnitude" % Filter)
plt.xlabel("Days since JD %.5f (%s)" % (JD0, Date0))
plt.grid(True)
plot_title = "%s_lc-all.pdf" % (config_file.split(".")[0])
plt.savefig(plot_title)
print("Saved light curve plot as %s" % plot_title)

# Separate plots for object
for k in range(nstar):
    mk_plot(
        k + 2,
        JD0,
        Date0,
        Objects[k],
        Ra_hms[k],
        Dec_dms[k],
        mjd_g,
        mjd_r,
        G_mag[k],
        G_mag_err[k],
        R_mag[k],
        R_mag_err[k],
        ywidth,
        line,
    )


# write CSV output file if requested
if csvfile != "":
    fn = open(csvfile, "w")
    for j in range(len(JD_G)):
        str1 = " ".join(
            "%s: %.3f %.3f" % (Objects[k], G_mag[k][j], G_mag_err[k][j])
            for k in range(nstar)
        )
        s = "%10.5f	 %30s	G  %s\n" % (JD_G[j], FitsFiles_all_G[j], str1)
        fn.write(s)
    for j in range(len(JD_R)):
        str1 = " ".join(
            "%s: %.3f %.3f" % (Objects[k], R_mag[k][j], R_mag_err[k][j])
            for k in range(nstar)
        )
        s = "%10.5f	 %30s	R  %s\n" % (JD_R[j], FitsFiles_all_R[j], str1)
        fn.write(s)
    print("Wrote data file: %s" % csvfile)
    fn.close()



import time

import pytest


def test_connected(device, disconnect):
    assert device.Connected
    device.Connected = False
    assert not device.Connected
    device.Connected = True
    assert device.Connected


def test_description(device):
    assert device.Description is not None


def test_driver_info(device):
    assert device.DriverInfo is not None


def test_driver_version(device):
    assert device.DriverVersion is not None


def test_interface_version(device):
    assert device.InterfaceVersion is not None


def test_name(device):
    assert device.Name is not None


def test_supported_actions(device):
    assert device.SupportedActions is not None


def test_start_exposure(device, disconnect):
    device.StartExposure(0.1, True)
    time.sleep(0.2)
    assert device.ImageReady


def test_stop_exposure(device, disconnect):
    if device.CanStopExposure:
        device.StartExposure(0.1, True)
        device.StopExposure()


def test_abort_exposure(device, disconnect):
    if device.CanAbortExposure:
        device.StartExposure(0.1, True)
        device.AbortExposure()


@pytest.mark.skip(reason="Not implemented")
def test_bayeroffsetx(device, disconnect):
    assert device.BayerOffsetX is not None


@pytest.mark.skip(reason="Not implemented")
def test_bayeroffsety(device, disconnect):
    assert device.BayerOffsetY is not None


def test_binx(device, disconnect):
    assert device.BinX is not None
    device.BinX = 2
    assert device.BinX == 2
    device.BinX = 1


def test_biny(device, disconnect):
    assert device.BinY is not None
    device.BinY = 2
    assert device.BinY == 2
    device.BinY = 1


def test_camera_state(device, disconnect):
    assert device.CameraState is not None


def test_cameraxsize(device, disconnect):
    assert device.CameraXSize is not None


def test_cameraysize(device, disconnect):
    assert device.CameraYSize is not None


def test_canasymmetricbin(device, disconnect):
    assert device.CanAsymmetricBin is not None


def test_ccdtemperature(device, disconnect):
    assert device.CCDTemperature is not None


def test_cooleron(device, disconnect):
    assert not device.CoolerOn
    device.CoolerOn = True
    assert device.CoolerOn
    device.CoolerOn = False
    assert not device.CoolerOn


def test_coolerpower(device, disconnect):
    if device.CanGetCoolerPower:
        assert device.CoolerPower is not None


def test_electronsperadu(device, disconnect):
    assert device.ElectronsPerADU is not None


def test_exposuremax(device, disconnect):
    assert device.ExposureMax is not None


def test_exposuremin(device, disconnect):
    assert device.ExposureMin is not None


def test_exposureresolution(device, disconnect):
    assert device.ExposureResolution is not None


def test_canfastreadout(device, disconnect):
    if device.CanFastReadout:
        assert device.FastReadout is not None


def test_fullwellcapacity(device, disconnect):
    assert device.FullWellCapacity is not None


@pytest.mark.skip(reason="Not implemented")
def test_gain(device, disconnect):
    if device.Gains is not None:
        device.Gain = 1
        assert device.Gain == 1
        assert device.GainMax is not None
        assert device.GainMin is not None
    else:
        device.Gain = device.Gains[0]
        assert device.Gain == device.Gains[0]


def test_hasshutter(device, disconnect):
    assert device.HasShutter is not None


@pytest.mark.skip(reason="Not implemented")
def test_heatsinktemperature(device, disconnect):
    assert device.HeatSinkTemperature is not None


def test_ispulseguiding(device, disconnect):
    if device.CanPulseGuide:
        assert device.IsPulseGuiding is not None


@pytest.mark.order(after="test_start_exposure")
def test_lastexposureduration(device, disconnect):
    device.StartExposure(0.1, True)
    while not device.ImageReady:
        time.sleep(0.1)
    assert device.ImageArray is not None
    assert device.LastExposureDuration is not None


@pytest.mark.order(after="test_start_exposure")
def test_lastexposurestarttime(device, disconnect):
    device.StartExposure(0.1, True)
    while not device.ImageReady:
        time.sleep(0.1)
    assert device.LastExposureStartTime is not None


def test_maxadu(device, disconnect):
    assert device.MaxADU is not None


def test_maxbinx(device, disconnect):
    assert device.MaxBinX is not None


def test_maxbiny(device, disconnect):
    assert device.MaxBinY is not None


def test_numx(device, disconnect):
    old_numx = device.NumX
    device.NumX = 2
    assert device.NumX == 2
    device.NumX = old_numx


def test_numy(device, disconnect):
    old_numy = device.NumY
    device.NumY = 2
    assert device.NumY == 2
    device.NumY = old_numy


@pytest.mark.skip(reason="Not implemented")
def test_offset(device, disconnect):
    if device.Offsets is not None:
        device.Offset = 1
        assert device.Offset == 1
        assert device.OffsetMax is not None
        assert device.OffsetMin is not None
    else:
        device.Offset = device.Offsets[0]
        assert device.Offset == device.Offsets[0]


def test_percentcompleted(device, disconnect):
    assert device.PercentCompleted is not None


def test_pixelsizex(device, disconnect):
    assert device.PixelSizeX is not None


def test_pixelsizey(device, disconnect):
    assert device.PixelSizeY is not None


def test_readoutmode(device, disconnect):
    assert device.ReadoutMode is not None
    device.ReadoutMode = 0
    assert device.ReadoutMode == 0


def test_sensorname(device, disconnect):
    assert device.SensorName is not None


def test_sensortype(device, disconnect):
    assert device.SensorType is not None


def test_setccdtemperature(device, disconnect):
    if device.CanSetCCDTemperature:
        device.SetCCDTemperature = -10
        assert device.SetCCDTemperature == -10


def test_startx(device, disconnect):
    assert device.StartX is not None
    device.StartX = 0
    assert device.StartX == 0


def test_starty(device, disconnect):
    assert device.StartY is not None
    device.StartY = 0
    assert device.StartY == 0


@pytest.mark.skip(reason="Not implemented")
def test_subexposureduration(device, disconnect):
    assert device.SubExposureDuration is not None
    device.SubExposureDuration = 1
    assert device.SubExposureDuration == 1


"""def test_properties(device, settings, disconn):
    print(settings)
    assert device.CameraXSize == settings["CameraXSize"]
    assert device.CameraYSize == settings["CameraYSize"]
    assert device.CanAbortExposure == settings["CanAbortExposure"]
    assert device.CanAsymmetricBin == settings["CanAsymmetricBin"]
    assert device.CanFastReadout == settings["CanFastReadout"]
    assert device.CanGetCoolerPower == settings["CanGetCoolerPower"]
    assert device.CanPulseGuide == settings["CanPulseGuide"]
    assert device.CanSetCCDTemperature == settings["CanSetCCDTemperature"]
    assert device.ElectronsPerADU == settings["ElectronsPerADU"]
    assert device.ExposureMax == settings["MaxExposure"]
    assert device.ExposureMin == settings["MinExposure"]
    assert device.ExposureResolution == settings["ExposureResolution"]
    # assert device.FastReadout == settings['FastReadout']
    assert device.FullWellCapacity == settings["FullWellCapacity"]
    assert device.HasShutter == settings["HasShutter"]
    assert device.MaxADU == settings["MaxADU"]
    assert device.MaxBinX == settings["MaxBinX"]
    assert device.MaxBinY == settings["MaxBinY"]
    assert device.PixelSizeX == settings["PixelSizeX"]
    assert device.PixelSizeY == settings["PixelSizeY"]
    v = settings["ReadoutModes"]
    assert device.ReadoutModes == v.split(",")
    assert device.SensorName == settings["SensorName"]
    assert device.SensorType == SensorType(settings["SensorType"])


def test_cooler(device, settings, disconn):
    # assert settings["HasCooler"], "OmniSim Has Cooler must be ON"
    assert device.CanSetCCDTemperature, "OmniSim Can Set CCD Temperature must be ON"
    assert device.CanGetCoolerPower, "OmniSim Can Get Cooler Power must be ON"


@pytest.mark.skipif(
    (c_sets.get("GainMode", None) != 1 or c_sets.get("OffsetMode", None) != 1),
    reason='Requires OmniSim Gain and Offset modes to be "Gain Names"',
)
def test_named_gains_offsets(device, settings, disconn):
    v = settings["Gains"]
    assert device.Gains == v.split(","), "OmniSim must be set to Named Gains mode"
    device.Gain = 0
    assert device.Gain == 0
    v = settings["Offsets"]
    assert device.Offsets == v.split(","), 'OmniSim must be set to "Named Offsets" mode'
    device.Offset = 0
    assert device.Offset == 0


@pytest.mark.skipif(
    (c_sets.get("GainMode", None) != 1 or c_sets.get("OffsetMode", None) != 1),
    reason='Requires OmniSim Gain and Offset modes to be "Min Max"',
)
def test_min_max_gains_offsets(device, settings, disconn):
    assert (
        device.GainMin == settings["GainMin"]
    ), "OmniSim must be set to Max-Min Gains mode"
    assert device.GainMax == settings["GainMax"]
    tg = device.GainMin + (device.GainMax - device.GainMin) // 2
    device.Gain = tg
    assert device.Gain == tg
    assert (
        device.OffsetMin == settings["OffsetMin"]
    ), "OmniSim must be set to Max-Min Offsets mode"
    assert device.OffsetMax == settings["OffsetMax"]
    tg = device.OffsetMin + (device.OffsetMax - device.OffsetMin) // 2
    device.Offset = tg
    assert device.Offset == tg


@pytest.mark.skipif(
    (c_sets.get("GainMode", None) != 1 or c_sets.get("OffsetMode", None) != 1),
    reason='Requires OmniSim Gain and Offset modes to be "None"',
)
def test_disabled_gains_offsets(device, settings, disconn):
    with pytest.raises(
        NotImplementedException, match=".*Gain[^s].*"
    ):  # Assure exactly "Gain"
        v = device.Gain
    with pytest.raises(
        NotImplementedException, match=".*Offset[^s].*"
    ):  # Assure exactly "Offset"
        v = device.Offset
    with pytest.raises(NotImplementedException, match=".*GainMax.*"):
        v = device.GainMax
    with pytest.raises(
        NotImplementedException, match=".*GainMax.*"
    ):  # Sim (0.1.2) returns error for GainMax
        v = device.GainMin
    with pytest.raises(NotImplementedException, match=".*Gains.*"):
        v = device.Gains
    with pytest.raises(NotImplementedException, match=".*OffsetMax.*"):
        v = device.OffsetMax
    with pytest.raises(
        NotImplementedException, match=".*OffsetMax.*"
    ):  # Sim (0.1.2) returns error for OffsetMax
        v = device.OffsetMin
    with pytest.raises(NotImplementedException, match=".*Offsets.*"):
        v = device.Offsets


@pytest.mark.skipif(
    (c_sets.get("SensorType", None) == 0),
    reason="Requires OmniSim SensorType to be other than Monochrome",
)
def test_color_bayer(device, settings, disconn):
    assert device.BayerOffsetX == settings["BayerOffsetX"]
    assert device.BayerOffsetY == settings["BayerOffsetY"]


@pytest.mark.skipif(
    (c_sets.get("SensorType", None) == 0),
    reason="Requires OmniSim SensorType to be Monochrome",
)
def test_image_capture(device, settings, disconn):
    assert (
        device.CameraXSize != device.CameraYSize
    ), "Width must not be the same as height for test validity"
    assert device.MaxBinX >= 2, "Camera must support X binning >= 2"
    assert device.MaxBinY >= 2, "Camera must support Y binning >= 2"
    device.BinX = 2
    assert device.BinX == 2
    device.BinY = 2
    assert device.BinY == 2
    device.StartX = 0
    assert device.StartX == 0
    device.StartY = 0
    assert device.StartY == 0
    device.NumX = device.CameraXSize // device.BinX
    assert device.NumX == device.CameraXSize // device.BinX
    device.NumY = device.CameraYSize // device.BinY
    assert device.NumY == device.CameraYSize // device.BinY
    device.StartExposure(1.0, True)
    while not device.ImageReady:
        time.sleep(0.5)
    img = device.ImageArray
    assert len(img) == device.NumX
    assert len(img[0]) == device.NumY


def test_image_stop_abort(device, settings, disconn):
    assert device.MaxBinX >= 2, "OmniSim Camera must support X binning >= 2"
    assert device.MaxBinY >= 2, "OmniSim OmniSim Camera must support Y binning >= 2"
    assert device.CanAbortExposure, "OmniSim Can Abort Exposure must be ON"
    assert device.CanStopExposure, "OmniSim Can Stop Exposure must be ON"
    device.StartExposure(3, True)
    time.sleep(1)
    device.StopExposure
    while not device.ImageReady:
        time.sleep(0.1)
    device.StartExposure(3, True)
    time.sleep(1)
    device.AbortExposure
    while not device.ImageReady:
        time.sleep(0.1)"""



import pytest

from pyscope.observatory import (
    WCS,
    Autofocus,
    Camera,
    CoverCalibrator,
    Device,
    Dome,
    FilterWheel,
    Focuser,
    ObservingConditions,
    Rotator,
    SafetyMonitor,
    Switch,
    Telescope,
)
from pyscope.observatory._docstring_inheritee import _DocstringInheritee


@pytest.mark.parametrize(
    "cls_name",
    [
        Autofocus,
        Camera,
        CoverCalibrator,
        Dome,
        Device,
        FilterWheel,
        Focuser,
        ObservingConditions,
        Rotator,
        SafetyMonitor,
        Switch,
        Telescope,
        WCS,
    ],
)
class TestAllAbstractClasses:
    def test_meta(self, cls_name):
        assert type(cls_name) is _DocstringInheritee

    def test_abstract(self, cls_name):
        with pytest.raises(TypeError):
            cls_name()

    '''def test_init_docstring(self, cls_name):
        assert cls_name.__doc__ is None
        assert (
            cls_name.__init__.__doc__
            != """Initialize self.  See help(type(self)) for accurate signature."""
        )'''

    """def test_func_docstring(self, cls_name):
        for name in cls_name.__dict__:
            if name != "__doc__":
                assert getattr(cls_name, name).__doc__ is not None"""


@pytest.mark.parametrize(
    "cls_name,methods",
    [
        (Autofocus, {"Run", "Abort", "__init__"}),
        (
            Camera,
            {
                "AbortExposure",
                "PulseGuide",
                "StartExposure",
                "StopExposure",
                "__init__",
                "BayerOffsetX",
                "BayerOffsetY",
                "BinX",
                "BinY",
                "CameraState",
                "CameraXSize",
                "CameraYSize",
                "CanAbortExposure",
                "CanAsymmetricBin",
                "CanFastReadout",
                "CanGetCoolerPower",
                "CanPulseGuide",
                "CanSetCCDTemperature",
                "CanStopExposure",
                "CCDTemperature",
                "CoolerOn",
                "CoolerPower",
                "ElectronsPerADU",
                "ExposureMax",
                "ExposureMin",
                "ExposureResolution",
                "FastReadout",
                "FullWellCapacity",
                "Gain",
                "GainMax",
                "GainMin",
                "Gains",
                "HasShutter",
                "HeatSinkTemperature",
                "ImageArray",
                "ImageReady",
                "IsPulseGuiding",
                "LastExposureDuration",
                "LastExposureStartTime",
                "MaxADU",
                "MaxBinX",
                "MaxBinY",
                "NumX",
                "NumY",
                "Offset",
                "OffsetMax",
                "OffsetMin",
                "Offsets",
                "PercentCompleted",
                "PixelSizeX",
                "PixelSizeY",
                "ReadoutMode",
                "ReadoutModes",
                "SensorName",
                "SensorType",
                "SetCCDTemperature",
                "StartX",
                "StartY",
                "SubExposureDuration",
            },
        ),
        (
            CoverCalibrator,
            {
                "CalibratorOff",
                "CalibratorOn",
                "CloseCover",
                "HaltCover",
                "OpenCover",
                "Brightness",
                "CalibratorState",
                "CoverState",
                "MaxBrightness",
                "__init__",
            },
        ),
        (
            Dome,
            {
                "AbortSlew",
                "CloseShutter",
                "FindHome",
                "OpenShutter",
                "Park",
                "SetPark",
                "SlewToAltitude",
                "SlewToAzimuth",
                "SyncToAzimuth",
                "Altitude",
                "AtHome",
                "AtPark",
                "Azimuth",
                "CanFindHome",
                "CanPark",
                "CanSetAltitude",
                "CanSetAzimuth",
                "CanSetPark",
                "CanSetShutter",
                "CanSlave",
                "CanSyncAzimuth",
                "ShutterStatus",
                "Slaved",
                "Slewing",
                "__init__",
            },
        ),
        (Device, {"Connected", "Name", "__init__"}),
        (FilterWheel, {"FocusOffsets", "Names", "Position", "__init__"}),
        (
            Focuser,
            {
                "Halt",
                "Move",
                "Absolute",
                "IsMoving",
                "MaxIncrement",
                "MaxStep",
                "Position",
                "StepSize",
                "TempComp",
                "TempCompAvailable",
                "Temperature",
                "__init__",
            },
        ),
        (
            ObservingConditions,
            {
                "Refresh",
                "SensorDescription",
                "TimeSinceLastUpdate",
                "AveragePeriod",
                "CloudCover",
                "DewPoint",
                "Humidity",
                "Pressure",
                "RainRate",
                "SkyBrightness",
                "SkyQuality",
                "SkyTemperature",
                "StarFWHM",
                "Temperature",
                "WindDirection",
                "WindGust",
                "WindSpeed",
                "__init__",
            },
        ),
        (
            Rotator,
            {
                "Halt",
                "Move",
                "MoveAbsolute",
                "MoveMechanical",
                "Sync",
                "CanReverse",
                "IsMoving",
                "MechanicalPosition",
                "Position",
                "Reverse",
                "StepSize",
                "TargetPosition",
                "__init__",
            },
        ),
        (SafetyMonitor, {"IsSafe", "__init__"}),
        (
            Switch,
            {
                "CanWrite",
                "GetSwitch",
                "GetSwitchDescription",
                "GetSwitchName",
                "MaxSwitchValue",
                "MinSwitchValue",
                "SetSwitch",
                "SetSwitchName",
                "SetSwitchValue",
                "SwitchStep",
                "MaxSwitch",
                "__init__",
            },
        ),
        (
            Telescope,
            {
                "AbortSlew",
                "AxisRates",
                "CanMoveAxis",
                "DestinationSideOfPier",
                "FindHome",
                "MoveAxis",
                "Park",
                "PulseGuide",
                "SetPark",
                "SlewToAltAz",
                "SlewToAltAzAsync",
                "SlewToCoordinates",
                "SlewToCoordinatesAsync",
                "SlewToTarget",
                "SlewToTargetAsync",
                "SyncToAltAz",
                "SyncToCoordinates",
                "SyncToTarget",
                "Unpark",
                "AlignmentMode",
                "Altitude",
                "ApertureArea",
                "ApertureDiameter",
                "AtHome",
                "AtPark",
                "Azimuth",
                "CanFindHome",
                "CanPark",
                "CanPulseGuide",
                "CanSetDeclinationRate",
                "CanSetGuideRates",
                "CanSetPark",
                "CanSetPierSide",
                "CanSetRightAscensionRate",
                "CanSetTracking",
                "CanSlew",
                "CanSlewAltAz",
                "CanSlewAltAzAsync",
                "CanSlewAsync",
                "CanSync",
                "CanSyncAltAz",
                "CanUnpark",
                "Declination",
                "DeclinationRate",
                "DoesRefraction",
                "EquatorialSystem",
                "FocalLength",
                "GuideRateDeclination",
                "GuideRateRightAscension",
                "IsPulseGuiding",
                "RightAscension",
                "RightAscensionRate",
                "SideOfPier",
                "SiderealTime",
                "SiteElevation",
                "SiteLatitude",
                "SiteLongitude",
                "Slewing",
                "SlewSettleTime",
                "TargetDeclination",
                "TargetRightAscension",
                "Tracking",
                "TrackingRate",
                "TrackingRates",
                "UTCDate",
                "__init__",
            },
        ),
        (WCS, {"Solve", "__init__"}),
    ],
)
class TestMethods:
    def test_methods(self, cls_name, methods):
        assert cls_name.__abstractmethods__ == frozenset(methods)



import logging

from .ascom_device import ASCOMDevice
from .filter_wheel import FilterWheel

logger = logging.getLogger(__name__)


class ASCOMFilterWheel(ASCOMDevice, FilterWheel):
    def __init__(self, identifier, alpaca=False, device_number=0, protocol="http"):
        super().__init__(
            identifier,
            alpaca=alpaca,
            device_type=self.__class__.__name__.replace("ASCOM", ""),
            device_number=device_number,
            protocol=protocol,
        )

    @property
    def FocusOffsets(self):
        logger.debug(f"ASCOMFilterWheel.FocusOffsets property called")
        return self._device.FocusOffsets

    @property
    def Names(self):
        logger.debug(f"ASCOMFilterWheel.Names property called")
        return self._device.Names

    @property
    def Position(self):
        logger.debug(f"ASCOMFilterWheel.Position property called")
        return self._device.Position

    @Position.setter
    def Position(self, value):
        logger.debug(f"ASCOMFilterWheel.Position property set to {value}")
        self._device.Position = value



import numpy as np


def airmass(alt):
    """Calculates the airmass given an altitude via Pickering 2002"""
    deg = np.pi / 180
    return 1 / np.sin((alt / deg + 244 / (165 + 47 * (alt / deg) ** 1.1)) * deg)



import pytest


def test_sensordescription(device, disconnect):
    assert device.SensorDescription("CloudCover") is not None
    assert device.SensorDescription("DewPoint") is not None
    assert device.SensorDescription("Humidity") is not None
    assert device.SensorDescription("Pressure") is not None
    assert device.SensorDescription("RainRate") is not None
    assert device.SensorDescription("SkyBrightness") is not None
    assert device.SensorDescription("SkyQuality") is not None
    assert device.SensorDescription("SkyTemperature") is not None
    assert device.SensorDescription("StarFWHM") is not None
    assert device.SensorDescription("Temperature") is not None
    assert device.SensorDescription("WindDirection") is not None
    assert device.SensorDescription("WindGust") is not None
    assert device.SensorDescription("WindSpeed") is not None


def test_timesincelastupdate(device, disconnect):
    device.Refresh()
    assert device.TimeSinceLastUpdate("CloudCover") is not None
    assert device.TimeSinceLastUpdate("DewPoint") is not None
    assert device.TimeSinceLastUpdate("Humidity") is not None
    assert device.TimeSinceLastUpdate("Pressure") is not None
    assert device.TimeSinceLastUpdate("RainRate") is not None
    assert device.TimeSinceLastUpdate("SkyBrightness") is not None
    assert device.TimeSinceLastUpdate("SkyQuality") is not None
    assert device.TimeSinceLastUpdate("SkyTemperature") is not None
    assert device.TimeSinceLastUpdate("StarFWHM") is not None
    assert device.TimeSinceLastUpdate("Temperature") is not None
    assert device.TimeSinceLastUpdate("WindDirection") is not None
    assert device.TimeSinceLastUpdate("WindGust") is not None
    assert device.TimeSinceLastUpdate("WindSpeed") is not None


def test_averageperiod(device, disconnect):
    assert device.AveragePeriod is not None
    # device.AveragePeriod = 0.1
    # assert device.AveragePeriod == 0.1
    # alpyca implementation issue


def test_cloudcover(device, disconnect):
    assert device.CloudCover is not None


def test_dewpoint(device, disconnect):
    assert device.DewPoint is not None


def test_humidity(device, disconnect):
    assert device.Humidity is not None


def test_pressure(device, disconnect):
    assert device.Pressure is not None


def test_rainrate(device, disconnect):
    assert device.RainRate is not None


def test_skybrightness(device, disconnect):
    assert device.SkyBrightness is not None


def test_skyquality(device, disconnect):
    assert device.SkyQuality is not None


def test_skytemperature(device, disconnect):
    assert device.SkyTemperature is not None


def test_starfwhm(device, disconnect):
    assert device.StarFWHM is not None


def test_temperature(device, disconnect):
    assert device.Temperature is not None


def test_winddirection(device, disconnect):
    assert device.WindDirection is not None


def test_windgust(device, disconnect):
    assert device.WindGust is not None


def test_windspeed(device, disconnect):
    assert device.WindSpeed is not None



#!/usr/bin/env python

# find-asteroid: Finds moving objects by comparing three images, removing fixed stars, and fitting for rectilinear motion in remaining objects
# v. 2.0 RLM Add MPC lookup

import os
import sys
from optparse import OptionParser

import numpy as np
import requests
import sewpy
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.table import Table

vers = "%prog 2.0 8-Dec-2016"


def get_args():
    global parser
    parser = OptionParser(
        description="%prog  finds moving object given exactly three input images",
        version=vers,
    )
    parser.add_option(
        "-f",
        dest="ftsfiles",
        metavar="FITS images",
        action="store",
        default="",
        help="comma-separated list of images [required]",
    )
    parser.add_option(
        "-m",
        dest="min_mag",
        metavar="Minimum mag.",
        action="store",
        type=float,
        default=21.0,
        help="Minimum magnitude [default 21.0]",
    )
    parser.add_option(
        "-v",
        dest="verbose",
        metavar="Verbose",
        action="store_true",
        default=False,
        help="Verbose output",
    )
    parser.add_option(
        "-r",
        dest="radius",
        default=15,
        metavar="Radius",
        action="store",
        help="Search radius [arcmin], default 15",
    )
    return parser.parse_args()


def hdr_info(fitsname):
    # Returns header info, substituting default values for zmag if needed
    hdr = fits.open(fitsname)[0].header
    filter = hdr["FILTER"][0]
    z = hdr["AIRMASS"]
    exptime = hdr["EXPTIME"]
    if filter == "G":
        k = 0.28
    elif filter == "R":
        k = 0.13
    elif filter == "W":
        k = 0.05
    elif filter == "N":
        k = 0.3
    elif filter == "B":
        k = 0.35
    else:
        k = 0.2
    if "ZMAG" in hdr:
        zmag = hdr["ZMAG"]
    else:
        if filter == "B":
            zmag = 20.0
        elif filter == "G":
            zmag = 21.05
        elif filter == "R":
            zmag = 20.20
        elif filter == "W":
            zmag = 20.65
        elif filter == "N":
            zmag = 21.8
        else:
            zmag = 20.0

        print(
            "WARNING: No Zero-point magnitude in header, using assumed zmag = %.2f , k = %.2f based on filter %s"
            % (zmag, k, filter)
        )
    return zmag, k, exptime, z


def get_hdr_info(ftsfile):
    im, hdr = fits.getdata(ftsfile, 0, header=True)
    object = hdr["OBJECT"]
    nbin = hdr["XBINNING"]
    D = hdr["DATE-OBS"]
    date = D[0:10]
    ut = D[11:]
    ra = hdr["RA"]
    dec = hdr["DEC"]
    exptime = hdr["EXPTIME"]
    filter = hdr["FILTER"]
    z = hdr["AIRMASS"]
    return object, nbin, date, ut, ra, dec, exptime, filter, z


def substring(string, i, j):
    return string[i:j]


def make2dlist(string):
    objlist = string.split("\n")
    object2d = []
    for index in range(len(objlist) - 1):
        object2d.append([])
        object2d[index].append(substring(str(objlist[index]), 0, 24))
        object2d[index].append(substring(str(objlist[index]), 24, 36))
        object2d[index].append(substring(str(objlist[index]), 36, 47))
        object2d[index].append(substring(str(objlist[index]), 47, 53))
        object2d[index].append(substring(str(objlist[index]), 53, 58))
        object2d[index].append(substring(str(objlist[index]), 59, 65))
        object2d[index].append(substring(str(objlist[index]), 69, 73))
        object2d[index].append(substring(str(objlist[index]), 76, 81))
        object2d[index].append(substring(str(objlist[index]), 82, 86))
        object2d[index].append(
            substring(str(objlist[index]), 87, len(objlist[index]) - 1)
        )
    del object2d[0:4]
    return object2d


def fix_signs(ra_off, dec_off, ra_dot, dec_dot):
    # Crack crazy suffixs N/S and +/- on MPC offsets, rates
    ra_off = ra_off.strip()
    dec_off = dec_off.strip()
    ra_dot = ra_dot.strip()
    dec_dot = dec_dot.strip()
    if ra_off.endswith("E"):
        ra_off = -1 * float(ra_off[:-1])
    else:
        ra_off = float(ra_off[:-1])
    if dec_off.endswith("S"):
        dec_off = -1 * float(dec_off[:-1])
    else:
        dec_off = float(dec_off[:-1])
    if ra_dot.endswith("-"):
        ra_dot = -1 * float(ra_dot[:-1])
    else:
        ra_dot = float(ra_dot[:-1])
    if dec_dot.endswith("-"):
        dec_dot = -1 * float(dec_dot[:-1])
    else:
        dec_dot = float(dec_dot[:-1])
    return ra_off, dec_off, ra_dot, dec_dot


def get_stars(fitsname, theta_max, min_mag):
    """
    Uses sewpy to retrieves star positions, fluxes from Table T,
    omitting any with flags != 0, theta > theta_max
    """
    zmag, k, exptime, z = hdr_info(fitsname)
    out = sew(fitsname)
    T = out["table"]  # this is an astropy table
    adu = T["FLUX_BEST"]
    adu = [x / exptime for x in adu]
    mag = zmag - 1.091 * np.log(adu) - z * k
    N = len(T)
    RA_deg = []
    Dec_deg = []
    Coords = []
    Radius = []
    Mag = []
    Flag = T["FLAGS"]
    radius = T["FLUX_RADIUS"]
    for n in range(N):
        if Flag[n] == 0 and radius[n] < theta_max and mag[n] < min_mag:
            ra_deg = T["ALPHA_J2000"][n]
            dec_deg = T["DELTA_J2000"][n]
            RA_deg.append(ra_deg)
            Dec_deg.append(dec_deg)
            Coords.append(SkyCoord(ra_deg, dec_deg, unit="deg"))
            Mag.append(mag[n])
            Radius.append(radius[n])
    RA_deg, Dec_deg, Coords, Mag, Radius = list(
        zip(*sorted(zip(RA_deg, Dec_deg, Coords, Mag, Radius)))
    )  # Sort on RA
    return RA_deg, Dec_deg, Coords, Mag, Radius


def chisq(ra, dec, t):
    # Calculates linear fit and chi-square to ra vs time, dec vs. time, assuming origin is at first point
    ra = np.array(ra)
    dec = np.array(dec)
    dt = t[1:] - t[0]
    ra_dot = (np.sum(ra[1:] * dt) - ra[0] * np.sum(dt)) / np.sum(dt**2)
    dec_dot = (np.sum(dec[1:] * dt) - dec[0] * np.sum(dt)) / np.sum(dt**2)
    ra_mod = ra[0] + ra_dot * (t - t[0])
    dec_mod = dec[0] + dec_dot * (t - t[0])
    chisq = (1.0 / (len(t) - 2)) * np.sum((ra_mod - ra) ** 2 + (dec_mod - dec) ** 2)
    return chisq, ra_dot, dec_dot


def compare_2fields(j, k):
    # Returns indices of stars in field j that are and are not in field k
    global Coords, Nstars, max_sepn
    ra0 = np.array([s.ra.deg for s in Coords[j]])
    dec0 = np.array([s.dec.deg for s in Coords[j]])
    ra1 = np.array([s.ra.deg for s in Coords[k]])
    dec1 = np.array([s.dec.deg for s in Coords[k]])
    indices_match = []
    indices_nomatch = []
    for i in range(Nstars[j]):
        dra = np.abs(ra1 - ra0[i])
        ddec = np.abs(dec1 - dec0[i])
        s = np.sqrt(dra**2 + ddec**2)
        found = np.any(s < max_sepn)
        if found:
            indices_match.append(i)
        else:
            indices_nomatch.append(i)
    return indices_match, indices_nomatch


def find_no_match():
    # Find indices of stars in fields  that are not in either of the other 2 fields
    dum, no_match01 = compare_2fields(0, 1)
    dum, no_match02 = compare_2fields(0, 2)
    no_match0 = list(set(no_match01).intersection(no_match02))
    dum, no_match10 = compare_2fields(1, 0)
    dum, no_match12 = compare_2fields(1, 2)
    no_match1 = list(set(no_match10).intersection(no_match12))
    dum, no_match20 = compare_2fields(2, 0)
    dum, no_match21 = compare_2fields(2, 1)
    no_match2 = list(set(no_match20).intersection(no_match21))
    return [no_match0, no_match1, no_match2]


def find_mpc_objects(ftsfile, radius, limmag):
    """Query the MPC database for small bodies within a specified radius [arcmin] and limiting magnitude of the center of a FITS image"""
    Objects = []
    Obj_Coords = []
    Magnitudes = []
    Offsets = []
    Rates = []
    Comments = []
    object, nbin, date, ut, ra, dec, exptime, filter, z = get_hdr_info(ftsfile)
    year, month, day = [int(s) for s in date.split("-")]
    uth, utm, uts = [float(s) for s in ut.split(":")]
    day += (uth + utm / 60.0 + uts / 3600.0) / 24.0
    day = "%.2f" % day
    coord = "%s %s" % (ra, dec)
    Ctr_Coords = SkyCoord(coord, unit=(u.hourangle, u.deg))
    ra = ra.replace(":", " ")
    dec = dec.replace(":", " ")
    payload = {
        "year": year,
        "month": month,
        "day": day,
        "which": "pos",
        "ra": ra,
        "decl": dec,
        "TextArea": " ",
        "radius": radius,
        "limit": limmag,
        "oc": "857",
        "sort": "d",
        "mot": "h",
        "tmot": "s",
        "pdes": "u",
        "needed": "f",
        "ps": "n",
        "type": "p",
    }
    r = requests.get(
        "http://www.minorplanetcenter.net/cgi-bin/mpcheck.cgi", params=payload
    )
    myString = r.text
    # check to see if any objects found
    if "No known minor planets" in r.text:
        return Objects, Ctr_Coords, Obj_Coords, Magnitudes, Offsets, Rates, Comments
    else:
        mySubString = r.text[r.text.find("<pre>") + 5 : r.text.find("</pre>")]
        mySubString = mySubString.replace("&#176;", "d")
        mySubString = mySubString.replace(
            '<a href="http://www.cfa.harvard.edu/iau/info/FurtherObs.html">Further observations?</a>',
            "",
        )
        # make a 2d list of objects for manipulation by other programs
        objectlist = make2dlist(mySubString)
        for line in objectlist:
            objname, ra, dec, mag, ra_off, dec_off, ra_dot, dec_dot, dum, comment = line
            ra_off, dec_off, ra_dot, dec_dot = fix_signs(
                ra_off, dec_off, ra_dot, dec_dot
            )
            mag = float(mag)
            Objects.append(objname)
            coord = "%s %s" % (ra, dec)
            c = SkyCoord(coord, unit=(u.hourangle, u.deg))
            Obj_Coords.append(c)
            Magnitudes.append(mag)
            Offsets.append([ra_off, dec_off])
            Rates.append([ra_dot, dec_dot])
            Comments.append(comment)
    return Objects, Ctr_Coords, Obj_Coords, Magnitudes, Offsets, Rates, Comments


# ======= MAIN =============

max_sepn = 1 / 3600.0  # Max separation for match, deg
theta_max = 1.0  # Set largest allowable star width [pixels

# Parse command line arguments
(opts, args) = get_args()
ftsfiles = opts.ftsfiles.split(",")
Nfts = len(ftsfiles)
if len(ftsfiles) != 3:
    sys.exit(
        "Exactly 3 FITS image names (comma separated) need to be given (option -f), try again"
    )
min_mag = opts.min_mag
verbose = opts.verbose
search_radius = opts.radius

# Define Sextractor dictionary items to retrieve
sew = sewpy.SEW(
    params=["ALPHA_J2000", "DELTA_J2000", "FLUX_RADIUS(3)", "FLUX_BEST", "FLAGS"],
    config={"DETECT_MINAREA": 5, "PHOT_FLUXFRAC": "0.3, 0.5, 0.7"},
)

# Loop through FITS files, getting lists of star positions and fluxes
Coords = []
Coords_hms = []
ADU = []
Nstars = []
Radius = []
Mag = []
jd = []
RA_deg = []
Dec_deg = []
j = 0

print("Sampling images for objects...")
for ftsfile in ftsfiles:
    if not os.path.isfile(ftsfile):
        sys.exit("File %s not found, exiting" % ftsfile)
    hdr = fits.open(ftsfile)[0].header
    jd.append(hdr["JD"])
    ra_deg, dec_deg, coords, mag, radius = get_stars(ftsfile, theta_max, min_mag)
    Radius.append(radius)
    Mag.append(mag)
    RA_deg.append(ra_deg)
    Dec_deg.append(dec_deg)
    Coords.append(coords)
    c = [
        s.to_string(style="hmsdms", precision=2, sep=":", decimal=False) for s in coords
    ]
    Coords_hms.append(c)
    Nstars.append(len(mag))
    print("%i stars found in FITS image %s" % (Nstars[j], ftsfile))
    j += 1
print()
print("Looking for moving objects...")
no_match = find_no_match()

if verbose:
    # Print all stars found in each field
    for k in range(3):
        print("Field %i stars" % k)
        for j in range(Nstars[k]):
            print("%s   %.2f" % (Coords_hms[k][j], Mag[k][j]))
        print()

    # Print star in each field that have no matching stars
    for k in range(3):
        print("Image: %s - stars with no matches" % ftsfiles[k])
        no_match[k].sort()
        for j in no_match[k]:
            print("%s   %.2f" % (Coords_hms[k][j], Mag[k][j]))

# Build lists of star coordinates [deg] for non-matching stars in each field
ra0 = []
dec0 = []
ra1 = []
dec1 = []
ra2 = []
dec2 = []
for k in no_match[0]:
    ra0.append(RA_deg[0][k])
    dec0.append(Dec_deg[0][k])
for k in no_match[1]:
    ra1.append(RA_deg[1][k])
    dec1.append(Dec_deg[1][k])
for k in no_match[2]:
    ra2.append(RA_deg[2][k])
    dec2.append(Dec_deg[2][k])

# Print solutions for any no-match stars sets that lie along a linear trajectory
print()
k_found = 0
ra_found = []
dec_found = []
for k0 in range(len(ra0)):
    for k1 in range(len(ra1)):
        for k2 in range(len(ra2)):
            ra = np.array([ra0[k0], ra1[k1], ra2[k2]])
            dec = np.array([dec0[k0], dec1[k1], dec2[k2]])
            t = np.array([jd[0], jd[1], jd[2]])
            chi, ra_dot, dec_dot = chisq(ra, dec, t)
            chi *= 1.0e6
            if chi < 0.1:
                print("Object nr. %i, unnormalized chisq = %.2e" % (k_found + 1, chi))
                print(
                    "Image %s:   %10.4f   %s    %.2f"
                    % (
                        ftsfiles[0],
                        jd[0],
                        Coords_hms[0][no_match[0][k0]],
                        Mag[0][no_match[0][k0]],
                    )
                )
                print(
                    "Image %s:   %10.4f   %s    %.2f"
                    % (
                        ftsfiles[1],
                        jd[1],
                        Coords_hms[1][no_match[1][k1]],
                        Mag[1][no_match[1][k1]],
                    )
                )
                print(
                    "Image %s:   %10.4f   %s    %.2f"
                    % (
                        ftsfiles[2],
                        jd[2],
                        Coords_hms[2][no_match[2][k2]],
                        Mag[2][no_match[2][k2]],
                    )
                )
                print(
                    'Motion: RA = %.1f"/hr, Dec = %.1f"/hr'
                    % (ra_dot * 3600 / 24.0, dec_dot * 3600 / 24.0)
                )
                ra_found.append(ra0[k0])
                dec_found.append(dec0[k0])
                k_found += 1
                print()

# Now perform MPC query of first field, report results
print("Looking for known objects using MPC online query...")
ftsfile = ftsfiles[0]
object, nbin, date, ut, ra, dec, exptime, filter, z = get_hdr_info(ftsfile)
(
    Objects,
    Ctr_Coords,
    Obj_Coords,
    Magnitudes,
    Offsets,
    Rates,
    Comments,
) = find_mpc_objects(ftsfile, search_radius, min_mag)
ctr_coords = Ctr_Coords.to_string(style="hmsdms", precision=1, sep=":", decimal=False)
ra_mpc = []
dec_mpc = []

# Print results
N = len(Objects)
if N == 0:
    print(
        "No known minor bodies with V < %s  within %s' radius of %s on %s at %s UT"
        % (min_mag, search_radius, ctr_coords, date, ut)
    )
else:
    print()
    print(
        "MPC query found %i minor bodies with V < %s  within %s' radius of field center (%s) on %s at %s UT"
        % (N, min_mag, search_radius, ctr_coords, date, ut)
    )
    print()
    print(
        " Number   Name               V       RA(J2000)    Dec(J2000)      Offsets            Motion        Comment"
    )
    print(
        "----------------------------------------------------------------------------------------------------------------------"
    )
    for j in range(N):
        coords_hmsdms = Obj_Coords[j].to_string(
            style="hmsdms", precision=1, sep=":", decimal=False
        )
        ra_mpc.append(Obj_Coords[j].ra.degree)
        dec_mpc.append(Obj_Coords[j].dec.degree)
        coords = Obj_Coords[j].to_string(
            style="hmsdms", precision=2, sep=":", decimal=False
        )
        ra_off, dec_off = Offsets[j]
        ra_rate, dec_rate = Rates[j]
        print(
            "%s   %.2f    %s   [%5.1f',%5.1f']   [%3i\"/hr, %3i\"/hr]   %s"
            % (
                Objects[j],
                Magnitudes[j],
                coords,
                ra_off,
                dec_off,
                ra_rate,
                dec_rate,
                Comments[j],
            )
        )
print()
ra_mpc = np.array(ra_mpc)
dec_mpc = np.array(dec_mpc)
max_sepn = 5 / 3600.0
print("Object matches")
print(
    "----------------------------------------------------------------------------------"
)
# Spin through found objects , print matches to MPC objects
if k_found > 0:
    for k in range(k_found):
        dra = np.abs(ra_found[k] - ra_mpc)
        ddec = np.abs(dec_found[k] - dec_mpc)
        s = np.sqrt(dra**2 + ddec**2)
        index = np.where(s < max_sepn)
        if index[0].size > 0:
            j = np.min(index)
            coords = Obj_Coords[j].to_string(
                style="hmsdms", precision=2, sep=":", decimal=False
            )
            print(
                "Object %i matches position of %s at %s"
                % (k + 1, Objects[j].strip(), coords)
            )
        else:
            print("Object %i has no MPC match (New discovery?)" % (k + 1))
else:
    print("No matches found")



import pytest


def test_getswitchdescription(device, disconnect):
    for i in range(device.MaxSwitch):
        assert device.GetSwitchDescription(i) is not None


def test_getswitchname(device, disconnect):
    for i in range(device.MaxSwitch):
        assert device.GetSwitchName(i) is not None


def test_setswitchname(device, disconnect):
    for i in range(device.MaxSwitch):
        device.SetSwitchName(i, "test")


def test_setswitch(device, disconnect):
    for i in range(device.MaxSwitch):
        if device.CanWrite(i):
            device.SetSwitch(i, True)
            assert device.GetSwitch(i)
        else:
            assert device.GetSwitch(i) is not None


@pytest.mark.skip(reason="Not implemented")
def test_setswitchvalue(device, disconnect):
    for i in range(device.MaxSwitch):
        for j in range(
            device.MinSwitchValue(i), device.MaxSwitchValue(i) + 1, device.SwitchStep(i)
        ):
            if device.CanWrite(i):
                device.SetSwitchValue(i, j)
                assert device.GetSwitchValue(i) == j
            else:
                assert device.GetSwitchValue(i) is not None



from ._args_kwargs_config import _kwargs_to_config
from ._function_synchronicity import _force_async, _force_sync
from ._get_image_source_catalog import _get_image_source_catalog
from ._html_line_parser import _get_number_from_line
from .airmass import airmass
from .pyscope_exception import PyscopeException

__all__ = [
    "airmass",
    "PyscopeException",
]



#!/usr/bin/python

# Rename and sequentially number files

import os
import re
import sys

if len(sys.argv) == 1:
    print(" ")
    print("Usage: file_renumber.py basename file1.ext file2.ext ...")
    print(" ")
    sys.exit("Renumbers sequential files \n")
elif len(sys.argv) >= 3:
    basename = sys.argv[1]
    infiles = sys.argv[2:]
else:
    print(" ")
    print("Usage: file_renumber.py basename file1.ext file2.ext ...")
    print(" ")
    sys.exit("Renumbers sequential files with a new basename\n")

# Set a clobber flag True so that images can be overwritten
# Otherwise set it False for safety

n = -1

for infile in infiles:
    n = n + 1

    # Create an output file name

    inbase = os.path.splitext(os.path.basename(infile))[0]
    inext = os.path.splitext(os.path.basename(infile))[1]
    val = re.findall(r"\d+", inbase)
    intval = int(float(val[0]))
    valtxt = "%04d" % (intval,)
    outfile = basename + valtxt + inext
    os.rename(infile, outfile)

exit()



# isort: skip_file

import logging

logger = logging.getLogger(__name__)

from ._docstring_inheritee import _DocstringInheritee

from .device import Device

from .autofocus import Autofocus
from .camera import Camera
from .cover_calibrator import CoverCalibrator
from .dome import Dome
from .filter_wheel import FilterWheel
from .focuser import Focuser
from .observing_conditions import ObservingConditions
from .rotator import Rotator
from .safety_monitor import SafetyMonitor
from .switch import Switch
from .telescope import Telescope
from .wcs import WCS

from .ascom_device import ASCOMDevice
from .ascom_camera import ASCOMCamera
from .ascom_cover_calibrator import ASCOMCoverCalibrator
from .ascom_dome import ASCOMDome
from .ascom_filter_wheel import ASCOMFilterWheel
from .ascom_focuser import ASCOMFocuser
from .ascom_observing_conditions import ASCOMObservingConditions
from .ascom_rotator import ASCOMRotator
from .ascom_safety_monitor import ASCOMSafetyMonitor
from .ascom_switch import ASCOMSwitch
from .ascom_telescope import ASCOMTelescope

from .astrometry_net_wcs import AstrometryNetWCS
from .html_observing_conditions import HTMLObservingConditions
from .html_safety_monitor import HTMLSafetyMonitor
from .ip_cover_calibrator import IPCoverCalibrator
from .maxim import Maxim
from .pinpoint_wcs import PinpointWCS

from .simulator_server import SimulatorServer

from .observatory_exception import ObservatoryException
from .observatory import Observatory


# from .platesolve2_wcs import Platesolve2WCS
# from .platesolve3_wcs import Platesolve3WCS
from .pwi_autofocus import PWIAutofocus

# from .skyx import SkyX

from .collect_calibration_set import collect_calibration_set

__all__ = [
    "ASCOMCamera",
    "ASCOMCoverCalibrator",
    "ASCOMDevice",
    "ASCOMDome",
    "ASCOMFilterWheel",
    "ASCOMFocuser",
    "ASCOMObservingConditions",
    "ASCOMRotator",
    "ASCOMSwitch",
    "ASCOMTelescope",
    "AstrometryNetWCS",
    "Autofocus",
    "Camera",
    "CoverCalibrator",
    "Device",
    "Dome",
    "FilterWheel",
    "Focuser",
    "HTMLObservingConditions",
    "HTMLSafetyMonitor",
    "IPCoverCalibrator",
    "Maxim",
    "ObservatoryException",
    "Observatory",
    "ObservingConditions",
    "PinpointWCS",
    "PWIAutofocus",
    "Rotator",
    "SafetyMonitor",
    "Switch",
    "Telescope",
    "WCS",
    "collect_calibration_set",
    "SimulatorServer",
]



import pytest

from pyscope.observatory import SimulatorServer


@pytest.mark.skip(reason="pytest_sessionfinish should end server")
@pytest.mark.order("last")
def test_shutdown():
    del pytest.simulator_server



import time

import pytest


def test_move(device, disconnect):
    device.Move(20000)
    while device.IsMoving:
        time.sleep(0.1)
    if device.Absolute:
        assert device.Position == 20000


def test_halt(device, disconnect):
    device.Move(20000)
    time.sleep(0.1)
    device.Halt()


def test_maxincrement(device, disconnect):
    assert device.MaxIncrement is not None


def test_maxstep(device, disconnect):
    assert device.MaxStep is not None


def test_stepsize(device, disconnect):
    assert device.StepSize is not None


def test_tempcomp(device, disconnect):
    if device.TempCompAvailable:
        assert device.TempComp is not None
        device.TempComp = True
        assert device.TempComp is True
        device.TempComp = False
        assert device.TempComp is False


def test_temperature(device, disconnect):
    assert device.Temperature is not None



"""
.. container::

  |License| |Zenodo| |PyPI Version| |PyPI Python Versions| |PyPI Downloads| |Astropy| |Donate|

`pyscope` is a pure-Python package for robotic scheduling, operation, and control of small
optical telescopes.

.. important::
  If you use `pyscope` for data collection or work
  presented in a publication or talk, please help the project by
  properly `acknowledging or citing </cite>`_ the package.

`pyscope` is an
`open-source <license>` project that provides a set of tools to rapidly and easily
control astronomical instrumentation. It is designed to be modular and extensible,
allowing users to easily add support for new devices and observatories.
`pyscope` is built on top of the
`ASCOM <https://ascom-standards.org/>`__ standard, but also provides support for
non-ASCOM devices. Users may also access their devices through third-party applications
such as `MaxIm DL <https://diffractionlimited.com/product/maxim-dl/>`__.

Observatories who use `pyscope` can take
advantage of the `~pyscope.telrun`
module, which provides a simple interface for fully-robotic observatory control.

`pyscope` is aiming to become an
`Astropy-affiliated package <https://www.astropy.org/affiliated/>`__.

Features
========
* Control observatory hardware with Python
* Support for both `ASCOM <https://ascom-standards.org/>`__ and non-ASCOM devices
* `~pyscope.observatory.Observatory` convenience methods like
  `~pyscope.observatory.Observatory.autofocus` and
  `~pyscope.observatory.Observatory.recenter`
* `~pyscope.telrun` module for fully-robotic operation of
  an observatory
* Basic data reduction tools like `~pyscope.reduction.avg_fits` and `~pyscope.reduction.ccd_calib`
* Simple analysis scripts like `~pyscope.analysis.calc_zmag`
* Powered by `Astropy <https://www.astropy.org/>`__,
  `Astropy-affiliated <https://www.astropy.org/affiliated/>`__
  packages, and `ASCOM <https://ascom-standards.org/>`__

.. |License| image:: https://img.shields.io/pypi/l/pyscope
    :target: https://pypi.org/project/pyscope/
    :alt: License

.. |Zenodo| image:: https://zenodo.org/badge/DOI/10.5281/zenodo.8403570.svg
    :target: https://doi.org/10.5281/zenodo.8403570
    :alt: Zenodo

.. |PyPI Version| image:: https://img.shields.io/pypi/v/pyscope
    :target: https://pypi.org/project/pyscope/
    :alt: PyPI Version

.. |PyPI Python Versions| image:: https://img.shields.io/pypi/pyversions/pyscope?logo=Python
    :target: https://pypi.org/project/pyscope/
    :alt: PyPI Python Versions

.. |PyPI Downloads| image:: https://img.shields.io/pypi/dm/pyscope?logo=python
    :target: https://pypi.org/project/pyscope/
    :alt: PyPI Downloads

.. |Astropy| image:: http://img.shields.io/badge/powered%20by-AstroPy-orange.svg?style=flat
    :target: http://www.astropy.org
    :alt: Powered by Astropy

.. |Donate| image:: https://img.shields.io/badge/Donate-to_pyscope-crimson
  :target: https://github.com/sponsors/WWGolay
  :alt: Donate

"""

# isort: skip_file

import logging

__version__ = "0.1.5"

from . import utils
from . import observatory
from . import telrun
from . import reduction
from . import analysis

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

__all__ = ["analysis", "observatory", "reduction", "telrun", "utils"]



import time

import pytest


def test_halt(device, disconnect):
    device.Move(5)
    device.Halt()


def test_moveabsolute(device, disconnect):
    device.MoveAbsolute(5)
    while device.IsMoving:
        time.sleep(0.1)


def test_movemechanical(device, disconnect):
    device.MoveMechanical(5)
    while device.IsMoving:
        time.sleep(0.1)


def test_sync(device, disconnect):
    device.Sync(5)


def test_position(device, disconnect):
    assert device.Position is not None


def test_reverse(device, disconnect):
    if device.CanReverse:
        device.Reverse = True
        assert device.Reverse == True
        device.Reverse = False
        assert device.Reverse == False


def test_mechanicalposition(device, disconnect):
    assert device.MechanicalPosition is not None


def test_stepsize(device, disconnect):
    assert device.StepSize is not None


def test_targetposition(device, disconnect):
    device.Move(5)
    while device.IsMoving:
        time.sleep(0.1)
    device.Move(5)
    assert device.TargetPosition == 15



import multiprocessing
import os
import platform
import signal
import subprocess
from tarfile import TarFile
from zipfile import ZipFile


class SimulatorServer:
    def __init__(self):
        if platform.system() == "Darwin":
            sys_name = "macos-x64"
            zip_type = ".zip"
        elif platform.system() == "Linux":
            sys_name = "linux"
            zip_type = ".tar.xz"

            if "arm" in platform.machine():
                sys_name += "-armhf"
            elif "aarch64" in platform.machine():
                sys_name += "-aarch64"
            elif "x86_64" in platform.machine():
                sys_name += "-x64"

        elif platform.system() == "Windows":
            sys_name = "windows"
            zip_type = ".zip"

            if "x86_64" in platform.machine():
                sys_name += "-x86"
            elif "x64" in platform.machine():
                sys_name += "-x64"

        else:
            raise Exception("Unsupported platform")

        dirname = "ascom.alpaca.simulators." + sys_name

        p = subprocess.check_output(
            f"""gh release download v0.3.1 --repo ASCOMInitiative/ASCOM.Alpaca.Simulators --skip-existing --pattern {dirname}{zip_type}""",
            shell=True,
        )

        if zip_type == ".zip":
            with ZipFile(dirname + zip_type, "r") as zipObj:
                zipObj.extractall()
        elif zip_type == ".tar.xz":
            with TarFile.open(dirname + zip_type, "r") as tarObj:
                tarObj.extractall()

        os.chmod(dirname + "/ascom.alpaca.simulators", 0o755)

        if platform.system() == "Darwin":
            self.process = subprocess.Popen(
                ("sudo " + dirname + "/ascom.alpaca.simulators"),
                preexec_fn=os.setpgrp,
                start_new_session=True,
                stderr=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                shell=True,
            )
        else:
            self.process = subprocess.Popen(
                (dirname + "/ascom.alpaca.simulators"),
                preexec_fn=os.setpgrp,
                start_new_session=True,
                stderr=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                shell=True,
            )

        self.dirname = dirname

    def __del__(self):
        subprocess.Popen(f"sudo kill {(os.getpgid(self.process.pid)+1)}", shell=True)



import datetime
import glob
import logging
import os
import shutil

import click
from astropy.io import fits

from ..analysis import calc_zmag
from ..observatory import AstrometryNetWCS
from .ccd_calib import ccd_calib

logger = logging.getLogger(__name__)


@click.command(
    epilog="""Check out the documentation at
                https://pyscope.readthedocs.io/ for more
                information."""
)
@click.option(
    "-t",
    "--camera-type",
    type=click.Choice(["ccd", "cmos"]),
    default="ccd",
    show_choices=True,
    show_default=True,
    help="Camera type.",
)
@click.option(
    "-i",
    "--image-dir",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    default="./images/",
    show_default=True,
    help="""Directory containing images to be calibrated. If passed, then
                    the fnames argument is ignored.""",
)
@click.option(
    "-c",
    "--calib-dir",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    default="./images/",
    show_default=True,
    required=True,
    help="Location of all calibration files.",
)
@click.option(
    "-r",
    "--raw-archive-dir",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    help="""Directory to archive raw images. If none given, no archiving is done,
                however, the '--in-place' option is not allowed.""",
)
@click.option(
    "--in-place",
    is_flag=True,
    default=False,
    show_default=True,
    help="""If given, the raw images are overwritten with the calibrated images.
                If not given, the calibrated images are written to the same directory as the
                raw images, but with the suffix '_cal' added to the filename prior to the extension.""",
)
@click.option(
    "-s",
    "--astro-scrappy",
    nargs=2,
    default=(1, 3),
    show_default=True,
    help="Number of hot pixel removal iterations and estimated camera read noise.",
)
@click.option(
    "-b",
    "--bad-columns",
    default="",
    show_default=True,
    help="Comma-separated list of bad columns to fix.",
)
@click.option(
    "-w",
    "--wcs",
    is_flag=True,
    default=False,
    show_default=True,
    help="""If given, the WCS is solved for each image. If not given, the WCS is
                not solved.""",
)
@click.option(
    "-z",
    "--zmag",
    is_flag=True,
    default=False,
    show_default=True,
    help="""If given, the zero-point magnitude is calculated for each image with
                SDSS filters. If not given, the zero-point magnitude is not calculated.""",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    show_default=True,
    help="Verbose output.",
)
@click.argument(
    "fnames", nargs=-1, type=click.Path(exists=True, file_okay=True, resolve_path=True)
)
@click.version_option()
def calib_images_cli(
    camera_type,
    image_dir,
    calib_dir,
    raw_archive_dir,
    in_place,
    astro_scrappy,
    bad_columns,
    wcs,
    zmag,
    verbose,
    fnames,
):
    if verbose:
        logger.setLevel(logging.DEBUG)

    if raw_archive_dir is None and in_place:
        raise click.BadParameter(
            "The --in-place option is not allowed if no raw archive directory is given."
        )

    if image_dir is not None:
        logger.info(
            "--image-dir passed, ignoring fnames argument and using all images in directory."
        )
        fnames = []
        for ext in (".fts", ".fits", ".fit"):
            fnames.extend(glob.glob(f"{image_dir}/*{ext}"))
        logger.debug(f"fnames = {fnames}")

    logger.info(f"Calibrating {len(fnames)} images.")
    for fname in fnames:
        logger.info(f"Calibrating {fname}")

        if raw_archive_dir is not None:
            raw_archive_dir = os.path.join(
                raw_archive_dir, datetime.datetime.now().strftime("%Y-%m-%d")
            )
            if not os.path.exists(raw_archive_dir):
                logger.info(f"Creating raw archive directory: {raw_archive_dir}")
                os.makedirs(raw_archive_dir)
            logger.info(f"Archiving {fname} to {raw_archive_dir}")
            shutil.copy(fname, raw_archive_dir)

        with fits.open(fname) as hdu:
            hdr = hdu[0].header

        if "CALSTAT" in hdr.keys():
            logger.info(f"{fname} has already been calibrated, skipping.")
            continue

        try:
            filt = hdr["FILTER"]
        except KeyError:
            filt = ""

        try:
            readout = hdr["READOUTM"]
        except KeyError:
            readout = hdr["READOUT"]

        try:
            exptime = round(hdr["EXPTIME"], 3)
        except KeyError:
            exptime = round(hdr["EXPOSURE"], 3)

        try:
            xbin = hdr["XBINNING"]
        except:
            xbin = hdr["XBIN"]

        try:
            ybin = hdr["YBINNING"]
        except:
            ybin = hdr["YBIN"]

        # TODO: Update these names, recursively search for latest set
        flat_frame = (
            f"{calib_dir}/master_flat_{filt}_{readout}_{exptime}_{xbin}x{ybin}.fts"
        )
        dark_frame = f"{calib_dir}/master_dark_{readout}_{exptime}_{xbin}x{ybin}.fts"

        if camera_type == "ccd":
            bias_frame = f"{calib_dir}/master_bias_{readout}_{xbin}x{ybin}.fts"
        elif camera_type == "cmos":
            flat_dark_frame = (
                f"{calib_dir}/master_flat_dark_{readout}_{exptime}_{xbin}x{ybin}.fts"
            )

        logger.debug("Using calibration frames:")
        logger.debug(f"Flat: {flat_frame}")
        logger.debug(f"Dark: {dark_frame}")
        if camera_type == "ccd":
            logger.debug(f"Bias: {bias_frame}")
        elif camera_type == "cmos":
            logger.debug(f"Flat dark: {flat_dark_frame}")

        logger.debug("Running ccd_calib...")
        ccd_calib(
            camera_type=camera_type,
            flat_frame=flat_frame,
            dark_frame=dark_frame,
            bias_frame=bias_frame,
            flat_dark_frame=flat_dark_frame,
            astro_scrappy=astro_scrappy,
            bad_columns=bad_columns,
            in_place=in_place,
            fnames=fname,
        )

        if wcs:
            logger.debug("Running Astrometry.net WCS solver...")
            solver = AstrometryNetWCS()
            solver.solve(fname)

        logger.debug("Done!")

    if zmag:
        logger.info("Calculating zero-point magnitudes...")
        calc_zmag(fnames=fnames)

    logger.info("Done!")


calib_images = calib_images_cli.callback



#!/usr/bin/env python

# Rocklister - lists MPC objects [using MPC database query] in a region and date extracted from a FITS header or specified by RA,Dec, UT Date/time

# Version 1.0 12/6/2016 RLM and Chris Michael
# 1.01 21 Jan 2017 update Optparser with usage string to show FITS image

vers = "rocklister v1.01"

import os
import sys
from optparse import OptionParser

import astropy.io.fits as fits
import requests
from astropy import units as u
from astropy.coordinates import SkyCoord


def get_args():
    parser = OptionParser(
        usage="Uasge: %prog [options] FITSimage",
        description="Program %prog queries MPC online database for Asteroids and Comets",
        version=vers,
    )
    parser.add_option(
        "-R",
        dest="RA",
        metavar="RA",
        action="store",
        help="RA to search [hh:mm:ss] (optional, default use FITS header)",
    )
    parser.add_option(
        "-D",
        dest="DEC",
        metavar="DEC",
        action="store",
        help="DEC to search [dd:mm] (optional, default use FITS header)",
    )
    parser.add_option(
        "-r",
        dest="radius",
        default=15,
        metavar="Radius",
        action="store",
        help="Search radius [arcmin], default 15",
    )
    parser.add_option(
        "-m",
        dest="limmag",
        default=20,
        metavar="LimMag",
        action="store",
        help="Limiting magnitude, default 20",
    )
    parser.add_option(
        "-d",
        dest="date",
        metavar="date",
        action="store",
        help="Date string [yyyy/mm/dd] (optional, default use FITS header)",
    )
    parser.add_option(
        "-t",
        dest="ut",
        metavar="time",
        action="store",
        default="00:00",
        help="UT time [hh:mm] (optional, default use FITS header)",
    )
    return parser.parse_args()


def get_hdr_info(ftsfile):
    im, hdr = fits.getdata(ftsfile, 0, header=True)
    object = hdr["OBJECT"]
    nbin = hdr["XBINNING"]
    D = hdr["DATE-OBS"]
    date = D[0:10]
    ut = D[11:]
    ra = hdr["RA"]
    dec = hdr["DEC"]
    exptime = hdr["EXPTIME"]
    filter = hdr["FILTER"]
    z = hdr["AIRMASS"]
    return object, nbin, date, ut, ra, dec, exptime, filter, z


def substring(string, i, j):
    return string[i:j]


def make2dlist(string):
    objlist = string.split("\n")
    object2d = []
    for index in range(len(objlist) - 1):
        object2d.append([])
        object2d[index].append(substring(str(objlist[index]), 0, 24))
        object2d[index].append(substring(str(objlist[index]), 24, 36))
        object2d[index].append(substring(str(objlist[index]), 36, 47))
        object2d[index].append(substring(str(objlist[index]), 47, 53))
        object2d[index].append(substring(str(objlist[index]), 53, 58))
        object2d[index].append(substring(str(objlist[index]), 59, 65))
        object2d[index].append(substring(str(objlist[index]), 69, 73))
        object2d[index].append(substring(str(objlist[index]), 76, 81))
        object2d[index].append(substring(str(objlist[index]), 82, 86))
        object2d[index].append(
            substring(str(objlist[index]), 87, len(objlist[index]) - 1)
        )
    del object2d[0:4]
    return object2d


def fix_signs(ra_off, dec_off, ra_dot, dec_dot):
    # Crack crazy suffixs N/S and +/- on MPC offsets, rates
    ra_off = ra_off.strip()
    dec_off = dec_off.strip()
    ra_dot = ra_dot.strip()
    dec_dot = dec_dot.strip()
    if ra_off.endswith("E"):
        ra_off = -1 * float(ra_off[:-1])
    else:
        ra_off = float(ra_off[:-1])
    if dec_off.endswith("S"):
        dec_off = -1 * float(dec_off[:-1])
    else:
        dec_off = float(dec_off[:-1])
    if ra_dot.endswith("-"):
        ra_dot = -1 * float(ra_dot[:-1])
    else:
        ra_dot = float(ra_dot[:-1])
    if dec_dot.endswith("-"):
        dec_dot = -1 * float(dec_dot[:-1])
    else:
        dec_dot = float(dec_dot[:-1])
    return ra_off, dec_off, ra_dot, dec_dot


def find_mpc_objects(ftsfile, radius, limmag):
    """Query the MPC database for small bodies within a specified radius [arcmin] and limiting magnitude of the center of a FITS image"""
    Objects = []
    Obj_Coords = []
    Magnitudes = []
    Offsets = []
    Rates = []
    Comments = []
    object, nbin, date, ut, ra, dec, exptime, filter, z = get_hdr_info(ftsfile)
    year, month, day = [int(s) for s in date.split("-")]
    uth, utm, uts = [float(s) for s in ut.split(":")]
    day += (uth + utm / 60.0 + uts / 3600.0) / 24.0
    day = "%.2f" % day
    coord = "%s %s" % (ra, dec)
    Ctr_Coords = SkyCoord(coord, unit=(u.hourangle, u.deg))
    ra = ra.replace(":", " ")
    dec = dec.replace(":", " ")
    payload = {
        "year": year,
        "month": month,
        "day": day,
        "which": "pos",
        "ra": ra,
        "decl": dec,
        "TextArea": " ",
        "radius": radius,
        "limit": limmag,
        "oc": "857",
        "sort": "d",
        "mot": "h",
        "tmot": "s",
        "pdes": "u",
        "needed": "f",
        "ps": "n",
        "type": "p",
    }
    r = requests.get(
        "http://www.minorplanetcenter.net/cgi-bin/mpcheck.cgi", params=payload
    )
    myString = r.text
    # check to see if any objects found
    if "No known minor planets" in r.text:
        return Objects, Ctr_Coords, Obj_Coords, Magnitudes, Offsets, Rates, Comments
    else:
        mySubString = r.text[r.text.find("<pre>") + 5 : r.text.find("</pre>")]
        mySubString = mySubString.replace("&#176;", "d")
        mySubString = mySubString.replace(
            '<a href="http://www.cfa.harvard.edu/iau/info/FurtherObs.html">Further observations?</a>',
            "",
        )
        # make a 2d list of objects for manipulation by other programs
        objectlist = make2dlist(mySubString)
        for line in objectlist:
            objname, ra, dec, mag, ra_off, dec_off, ra_dot, dec_dot, dum, comment = line
            ra_off, dec_off, ra_dot, dec_dot = fix_signs(
                ra_off, dec_off, ra_dot, dec_dot
            )
            mag = float(mag)
            Objects.append(objname)
            coord = "%s %s" % (ra, dec)
            c = SkyCoord(coord, unit=(u.hourangle, u.deg))
            Obj_Coords.append(c)
            Magnitudes.append(mag)
            Offsets.append([ra_off, dec_off])
            Rates.append([ra_dot, dec_dot])
            Comments.append(comment)
    return Objects, Ctr_Coords, Obj_Coords, Magnitudes, Offsets, Rates, Comments


def find_mpc_objects_cl(date, ut, ra, dec, radius, limmag):
    """Query the MPC database for small bodies within a specified radius [arcmin] and limiting magnitude using user-supplied ra,dec,date"""
    Objects = []
    Obj_Coords = []
    Magnitudes = []
    Offsets = []
    Rates = []
    Comments = []
    year, month, day = date.split("/")
    ut_hr, ut_min = [float(x) for x in ut.split(":")]
    frac_day = (ut_hr + ut_min / 60.0) / 24.0
    day = "%.2f" % (float(day) + frac_day)
    coord = "%s %s" % (ra, dec)
    Ctr_Coords = SkyCoord(coord, unit=(u.hourangle, u.deg))
    ra = ra.replace(":", " ")
    dec = dec.replace(":", " ")
    payload = {
        "year": year,
        "month": month,
        "day": day,
        "which": "pos",
        "ra": ra,
        "decl": dec,
        "TextArea": " ",
        "radius": radius,
        "limit": limmag,
        "oc": "857",
        "sort": "d",
        "mot": "h",
        "tmot": "s",
        "pdes": "u",
        "needed": "f",
        "ps": "n",
        "type": "p",
    }
    r = requests.get(
        "http://www.minorplanetcenter.net/cgi-bin/mpcheck.cgi", params=payload
    )
    myString = r.text
    # check to see if any objects found
    if "No known minor planets" in r.text:
        return Objects, Ctr_Coords, Obj_Coords, Magnitudes, Offsets, Rates, Comments
    else:
        mySubString = r.text[r.text.find("<pre>") + 5 : r.text.find("</pre>")]
        mySubString = mySubString.replace("&#176;", "d")
        mySubString = mySubString.replace(
            '<a href="http://www.cfa.harvard.edu/iau/info/FurtherObs.html">Further observations?</a>',
            "",
        )
        # make a 2d list of objects for manipulation by other programs
        objectlist = make2dlist(mySubString)
        for line in objectlist:
            objname, ra, dec, mag, ra_off, dec_off, ra_dot, dec_dot, dum, comment = line
            ra_off, dec_off, ra_dot, dec_dot = fix_signs(
                ra_off, dec_off, ra_dot, dec_dot
            )
            mag = float(mag)
            Objects.append(objname)
            coord = "%s %s" % (ra, dec)
            c = SkyCoord(coord, unit=(u.hourangle, u.deg))
            Obj_Coords.append(c)
            Magnitudes.append(mag)
            Offsets.append([ra_off, dec_off])
            Rates.append([ra_dot, dec_dot])
            Comments.append(comment)
    return Objects, Ctr_Coords, Obj_Coords, Magnitudes, Offsets, Rates, Comments


# ====== MAIN ==========

# Get params from command line
(opts, args) = get_args()
radius = opts.radius
limmag = opts.limmag


# Query MPC, parse output
if len(args) == 1:
    ftsfile = args[0]
    if os.path.isfile(ftsfile):
        (
            Objects,
            Ctr_Coords,
            Obj_Coords,
            Magnitudes,
            Offsets,
            Rates,
            Comments,
        ) = find_mpc_objects(ftsfile, radius, limmag)
        ctr_coords = Ctr_Coords.to_string(
            style="hmsdms", precision=1, sep=":", decimal=False
        )
        object, nbin, date, ut, ra, dec, exptime, filter, z = get_hdr_info(ftsfile)
    else:
        sys.exit("File %s does not exist in current path, exiting" % ftsfile)
else:
    if not opts.RA or not opts.DEC or not opts.date:
        sys.exit("Must specifiy at least -R, -D, -d if no image specified.")
    ra = opts.RA
    dec = opts.DEC
    ut = opts.ut
    date = opts.date
    (
        Objects,
        Ctr_Coords,
        Obj_Coords,
        Magnitudes,
        Offsets,
        Rates,
        Comments,
    ) = find_mpc_objects_cl(date, ut, ra, dec, radius, limmag)
    ctr_coords = Ctr_Coords.to_string(
        style="hmsdms", precision=1, sep=":", decimal=False
    )
# Print results
N = len(Objects)
if N == 0:
    print(
        "No known minor bodies with V < %s  within %s' radius of %s on %s at %s UT"
        % (limmag, radius, ctr_coords, date, ut)
    )
else:
    print()
    print(
        "Found %i  minor bodies with V < %s  within %s' radius of %s on %s at %s UT"
        % (N, limmag, radius, ctr_coords, date, ut)
    )
    print()
    print(
        " Number   Name               V       RA(J2000)    Dec(J2000)      Offsets            Motion        Comment"
    )
    print(
        "----------------------------------------------------------------------------------------------------------------------"
    )
    for j in range(N):
        coords_hmsdms = Obj_Coords[j].to_string(
            style="hmsdms", precision=1, sep=":", decimal=False
        )
        ra_deg = Obj_Coords[j].ra.degree
        dec_deg = Obj_Coords[j].dec.degree
        coords = Obj_Coords[j].to_string(
            style="hmsdms", precision=2, sep=":", decimal=False
        )
        ra_off, dec_off = Offsets[j]
        ra_rate, dec_rate = Rates[j]
        print(
            "%s   %.2f    %s   [%5.1f',%5.1f']   [%3i\"/hr, %3i\"/hr]   %s"
            % (
                Objects[j],
                Magnitudes[j],
                coords,
                ra_off,
                dec_off,
                ra_rate,
                dec_rate,
                Comments[j],
            )
        )



from abc import ABC, abstractmethod

from ._docstring_inheritee import _DocstringInheritee


class Device(ABC, metaclass=_DocstringInheritee):
    @abstractmethod
    def __init__(self, *args, **kwargs):
        pass

    @property
    @abstractmethod
    def Connected(self):
        pass

    @Connected.setter
    @abstractmethod
    def Connected(self, value):
        pass

    @property
    @abstractmethod
    def Name(self):
        pass



import pytest

from pyscope.observatory import (
    ASCOMCamera,
    ASCOMCoverCalibrator,
    ASCOMDevice,
    ASCOMDome,
    ASCOMFilterWheel,
    ASCOMFocuser,
    ASCOMObservingConditions,
    ASCOMRotator,
    ASCOMSafetyMonitor,
    ASCOMSwitch,
    ASCOMTelescope,
    Camera,
    CoverCalibrator,
    Device,
    Dome,
    FilterWheel,
    Focuser,
    ObservingConditions,
    Rotator,
    SafetyMonitor,
    SimulatorServer,
    Switch,
    Telescope,
)
from pyscope.observatory._docstring_inheritee import _DocstringInheritee


class TestASCOMDriver:
    def test_meta(self):
        assert type(ASCOMDevice) is _DocstringInheritee

    def test_parents(self):
        assert issubclass(ASCOMDevice, Device)


@pytest.mark.parametrize(
    "cls_name",
    [
        ASCOMCamera,
        ASCOMCoverCalibrator,
        ASCOMDome,
        ASCOMFilterWheel,
        ASCOMFocuser,
        ASCOMObservingConditions,
        ASCOMRotator,
        ASCOMSafetyMonitor,
        ASCOMSwitch,
        ASCOMTelescope,
    ],
)
class TestAllASCOMClasses:
    def test_meta(self, cls_name):
        assert type(cls_name) is _DocstringInheritee

    def test_driver_parent(self, cls_name):
        assert issubclass(cls_name, ASCOMDevice)

    def test_hardware_parent(self, cls_name):
        assert issubclass(cls_name, eval(cls_name.__name__.replace("ASCOM", "")))

    """def test_docstrings(self, cls_name):
        for name in cls_name.__dict__:
            if name != "__doc__":
                assert getattr(ASCOMDevice, name).__doc__ is not None"""



import logging
import time

import pytest

from pyscope import logger
from pyscope.observatory import ASCOMCamera, Observatory


def test_observatory(tmp_path):
    logger.setLevel("INFO")
    logger.addHandler(logging.StreamHandler())

    obs = Observatory(config_path="tests/reference/simulator_observatory.cfg")
    obs.save_config("tests/reference/saved_observatory.cfg")
    new_obs = Observatory(config_path="tests/reference/saved_observatory.cfg")
    obs.connect_all()
    obs.dome.CloseShutter()
    obs.dome.OpenShutter()

    assert obs.lst() is not None
    assert obs.sun_altaz() is not None
    assert obs.moon_altaz() is not None
    assert obs.moon_illumination() is not None
    assert obs.get_object_altaz(obj="M1") is not None
    assert obs.get_object_slew(obj="M1") is not None

    obs.start_observing_conditions_thread(update_interval=1)
    # obs.start_safety_monitor_thread()

    obs.telescope.Unpark()
    if obs.telescope.CanFindHome:
        obs.telescope.FindHome()
        assert obs.get_current_object() is not None

    obs.slew_to_coordinates(ra=obs.lst(), dec=45)
    obs.run_autofocus(midpoint=20000, exposure=1, use_current_pointing=True)

    old_position = obs.focuser.Position
    obs.set_filter_offset_focuser(filter_index=5)
    assert obs.filter_wheel.Position == 5
    assert obs.focuser.Position != old_position
    obs.start_derotation_thread()
    obs.camera.StartExposure(0.1, True)
    while not obs.camera.ImageReady:
        time.sleep(0.1)
    obs.save_last_image(
        str(tmp_path) + "last_image.fts", frametyp="light", do_fwhm=True, overwrite=True
    )
    obs.stop_derotation_thread()

    assert obs.safety_status() is not None
    assert obs.switch_status() is not None

    obs.take_flats(
        filter_exposure=6 * [0.1],
        filter_brightness=6 * [1],
        readouts=[0],
        repeat=1,
        save_path=str(tmp_path),
    )

    obs.take_darks(
        exposures=[0.1, 0.2],
        readouts=[0],
        binnings=["1x1"],
        repeat=1,
        save_path=str(tmp_path),
    )

    # obs.stop_safety_monitor_thread()
    obs.stop_observing_conditions_thread()
    obs.telescope.Park()
    obs.shutdown()
    obs.disconnect_all()


"""if __name__ == "__main__":
    test_observatory(tmp_path='tests/reference/')"""



import inspect
import os
import pathlib
import subprocess
import sys
from urllib.parse import quote

from packaging.version import parse
from sphinx_astropy.conf.v2 import *

sys.path.insert(0, pathlib.Path(__file__).parents[2].resolve().as_posix())

import pyscope

project = "pyscope"
copyright = "2023, Walter Golay"
author = "Walter Golay"
version = pyscope.__version__
release = version

html_theme = "pydata_sphinx_theme"

html_static_path = [os.path.join(os.path.dirname(__file__), "_static")]

graphviz_dot = "/usr/local/bin/dot"

html_theme_options = {
    "logo": {
        "image_light": "_static/pyscope_logo_small_gray.png",
        "image_dark": "_static/pyscope_logo_small.png",
    },
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/WWGolay/pyscope",
            "icon": "fa-brands fa-square-github",
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/pyscope/",
            "icon": "fa-brands fa-python",
        },
        {
            "name": "macro",
            "url": "https://macroconsortium.org",
            "icon": "fa-solid fa-shuttle-space",
        },
    ],
}

intersphinx_mapping["click"] = ("https://click.palletsprojects.com/en/8.1.x/", None)
intersphinx_mapping["astroquery"] = (
    "https://astroquery.readthedocs.io/en/latest/",
    None,
)
intersphinx_mapping["astroplan"] = ("https://astroplan.readthedocs.io/en/latest/", None)

extensions = list(map(lambda x: x.replace("viewcode", "linkcode"), extensions))


def linkcode_resolve(domain, info):
    """
    Determine the URL corresponding to Python object.

    Inspired by:
    - https://gist.github.com/nlgranger/55ff2e7ff10c280731348a16d569cb73
    - https://github.com/matplotlib/matplotlib/blob/main/doc/conf.py
    """
    if domain != "py":
        return None

    modname = info["module"]
    fullname = info["fullname"]

    submod = sys.modules.get(modname)
    if submod is None:
        return None

    obj = submod
    for part in fullname.split("."):
        try:
            obj = getattr(obj, part)
        except AttributeError:
            return None

    if inspect.isfunction(obj):
        obj = inspect.unwrap(obj)
    try:
        fn = inspect.getsourcefile(obj)
    except TypeError:
        fn = None
    if not fn or fn.endswith("__init__.py"):
        try:
            fn = inspect.getsourcefile(sys.modules[obj.__module__])
        except (TypeError, AttributeError, KeyError):
            fn = None
    if not fn:
        return None

    try:
        source, lineno = inspect.getsourcelines(obj)
    except (OSError, TypeError):
        lineno = None

    linespec = f"#L{lineno:d}-L{lineno + len(source) - 1:d}" if lineno else ""

    startdir = pathlib.Path(pyscope.__file__).parent.parent
    try:
        fn = os.path.relpath(fn, start=startdir).replace(os.path.sep, "/")
    except ValueError:
        return None

    linkcode_revision = "main"
    try:
        # lock to commit number
        cmd = "git log -n1 --pretty=%H"
        head = subprocess.check_output(cmd.split()).strip().decode("utf-8")
        linkcode_revision = head

        # if we are on main's HEAD, use main as reference
        cmd = "git log --first-parent main -n1 --pretty=%H"
        main = subprocess.check_output(cmd.split()).strip().decode("utf-8")
        if head == main:
            linkcode_revision = "main"

        # if we have a tag, use tag as reference
        cmd = "git describe --exact-match --tags " + head
        tag = (
            subprocess.check_output(cmd.split(" "), stderr=subprocess.DEVNULL)
            .strip()
            .decode("utf-8")
        )
        linkcode_revision = tag

    except subprocess.CalledProcessError:
        pass

    return (
        "https://github.com/WWGolay/pyscope/blob" f"/{linkcode_revision}/{fn}{linespec}"
    )


extensions.append("sphinx_favicon")

favicons = [
    "logo16.png",
    "logo32.png",
    "logo48.png",
    "logo128.png",
    "logo256.png",
    "pyscopeIcon.svg",
    {"rel": "apple-touch-icon", "href": "logo180.png"},
]

extensions.append("sphinx.ext.doctest")

# extensions.append("sphinxcontrib.programoutput")
# extensions.append("sphinx_exec_code")



import logging
import platform
import time

from .autofocus import Autofocus
from .camera import Camera
from .device import Device
from .filter_wheel import FilterWheel
from .wcs import WCS

logger = logging.getLogger(__name__)


class Maxim(Device):
    def __init__(self):
        logger.debug("Maxim.Maxim __init__ called")
        if platform.system() != "Windows":
            raise Exception("This class is only available on Windows.")
        else:
            from win32com.client import Dispatch

            self._app = Dispatch("MaxIm.Application")
            self._app.LockApp = True

            self._camera = _MaximCamera()
            self._camera.Connected = True

            self._filter_wheel = None
            self._autofocus = None
            if self._camera.HasFilterWheel:
                self._filter_wheel = _MaximFilterWheel(self.camera)
            if self._app.FocuserConnected:
                self._autofocus = _MaximAutofocus(self.app)

            self._wcs = _MaximPinpointWCS(self.camera)

    @property
    def Connected(self):
        logger.debug("Maxim.Connected called")
        return self.camera.Connected

    @Connected.setter
    def Connected(self, value):
        logger.debug(f"Connected setter called with value={value}")
        self.camera.Connected = value

    @property
    def Name(self):
        logger.debug("Maxim.Name called")
        return "MaxIm DL"

    @property
    def app(self):
        logger.debug("Maxim.app called")
        return self._app

    @property
    def autofocus(self):
        logger.debug("Maxim.autofocus called")
        return self._autofocus

    @property
    def camera(self):
        logger.debug("Maxim.camera called")
        return self._camera


class _MaximAutofocus(Autofocus):
    def __init__(self, maxim):
        logger.debug("_MaximAutofocus.MaximAutofocus __init__ called")
        self.maxim = maxim

    def Run(self, exposure=10):
        logger.debug(f"Run called with exposure={exposure}")
        self.maxim.Autofocus(exposure)

        while self.maxim.AutofocusStatus == -1:
            time.sleep(1)

        if self.maxim.AutofocusStatus == 1:
            return True
        else:
            return False

    def Abort(self):
        logger.debug("_MaximAutofocus.Abort called")
        raise NotImplementedError


class _MaximCamera(Camera):
    def __init__(self):
        logger.debug("_MaximCamera_MaximCamera __init__ called")
        self._com_object = Dispatch("MaxIm.CCDCamera")
        self._com_object.DisableAutoShutdown = True

        self._last_exposure_duration = None
        self._last_exposure_start_time = None

    @property
    def Connected(self):
        logger.debug("_MaximCameraConnected called")
        return self._com_object is not None and self._com_object.LinkEnabled

    @Connected.setter
    def Connected(self, value):
        logger.debug(f"Connected setter called with value={value}")
        self._com_object.LinkEnabled = value

    @property
    def Name(self):
        logger.debug("_MaximCameraName called")
        return self._com_object.CameraName

    def AbortExposure(self):
        logger.debug("_MaximCameraAbortExposure called")
        self._com_object.AbortExposure()

    def PulseGuide(self, Direction, Duration):
        logger.debug(
            f"PulseGuide called with Direction={Direction}, Duration={Duration}"
        )
        raise NotImplementedError

    def StartExposure(self, Duration, Light):
        logger.debug(f"StartExposure called with Duration={Duration}, Light={Light}")
        self._last_exposure_duration = Duration
        self._last_exposure_start_time = time.time()
        self._com_object.Expose(Duration, Light)

    def StopExposure(self):
        logger.debug("_MaximCameraStopExposure called")
        self._com_object.AbortExposure()

    @property
    def BayerOffsetX(self):
        logger.debug("_MaximCameraBayerOffsetX called")
        raise NotImplementedError

    @property
    def BayerOffsetY(self):
        logger.debug("_MaximCameraBayerOffsetY called")
        raise NotImplementedError

    @property
    def BinX(self):
        logger.debug("_MaximCameraBinX called")
        return self._com_object.BinX

    @BinX.setter
    def BinX(self, value):
        logger.debug(f"BinX setter called with value={value}")
        self._com_object.BinX = value

    @property
    def BinY(self):
        logger.debug("_MaximCameraBinY called")
        return self._com_object.BinY

    @BinY.setter
    def BinY(self, value):
        logger.debug(f"BinY setter called with value={value}")
        self._com_object.BinY = value

    @property
    def CameraState(self):
        logger.debug("_MaximCameraCameraState called")
        return self._com_object.CameraStatus

    @property
    def CameraXSize(self):
        logger.debug("_MaximCameraCameraXSize called")
        return self._com_object.CameraXSize

    @property
    def CameraYSize(self):
        logger.debug("_MaximCameraCameraYSize called")
        return self._com_object.CameraYSize

    @property
    def CanAbortExposure(self):
        logger.debug("_MaximCameraCanAbortExposure called")
        return True

    @property
    def CanAsymmetricBin(self):
        logger.debug("_MaximCameraCanAsymmetricBin called")
        return self._com_object.XYBinning

    @property
    def CanFastReadout(self):
        logger.debug("_MaximCameraCanFastReadout called")
        raise NotImplementedError

    @property
    def CanGetCoolerPower(self):
        logger.debug("_MaximCameraCanGetCoolerPower called")
        return True

    @property
    def CanPulseGuide(self):
        logger.debug("_MaximCameraCanPulseGuide called")
        return False

    @property
    def CanSetCCDTemperature(self):
        logger.debug("_MaximCameraCanSetCCDTemperature called")
        return self._com_object.CanSetTemperature

    @property
    def CanStopExposure(self):
        logger.debug("_MaximCameraCanStopExposure called")
        return True

    @property
    def CCDTemperature(self):
        logger.debug("_MaximCameraCCDTemperature called")
        return self._com_object.Temperature

    @property
    def CoolerOn(self):
        logger.debug("_MaximCameraCoolerOn called")
        return self._com_object.CoolerOn

    @CoolerOn.setter
    def CoolerOn(self, value):
        logger.debug(f"CoolerOn setter called with value={value}")
        self._com_object.CoolerOn = value

    @property
    def CoolerPower(self):
        logger.debug("_MaximCameraCoolerPower called")
        return self._com_object.CoolerPower

    @property
    def ElectronsPerADU(self):
        logger.debug("_MaximCameraElectronsPerADU called")
        raise NotImplementedError

    @property
    def ExposureMax(self):
        logger.debug("_MaximCameraExposureMax called")
        raise NotImplementedError

    @property
    def ExposureMin(self):
        logger.debug("_MaximCameraExposureMin called")
        raise NotImplementedError

    @property
    def ExposureResolution(self):
        logger.debug("_MaximCameraExposureResolution called")
        raise NotImplementedError

    @property
    def FastReadout(self):
        logger.debug("_MaximCameraFastReadout called")
        raise DeprecationWarning

    @FastReadout.setter
    def FastReadout(self, value):
        logger.debug(f"FastReadout setter called with value={value}")
        raise DeprecationWarning

    @property
    def FullWellCapacity(self):
        logger.debug("_MaximCameraFullWellCapacity called")
        raise NotImplementedError

    @property
    def Gain(self):
        logger.debug("_MaximCameraGain called")
        return self._com_object.Speed

    @Gain.setter
    def Gain(self, value):
        logger.debug(f"Gain setter called with value={value}")
        self._com_object.Speed = value

    @property
    def GainMax(self):
        logger.debug("_MaximCameraGainMax called")
        raise NotImplementedError

    @property
    def GainMin(self):
        logger.debug("_MaximCameraGainMin called")
        raise NotImplementedError

    @property
    def Gains(self):
        logger.debug("_MaximCameraGains called")
        return self._com_object.Speeds

    @property
    def HasShutter(self):
        logger.debug("_MaximCameraHasShutter called")
        return self._com_object.HasShutter

    @property
    def HeatSinkTemperature(self):
        logger.debug("_MaximCameraHeatSinkTemperature called")
        return self._com_object.AmbientTemperature

    @property
    def ImageArray(self):
        logger.debug("_MaximCameraImageArray called")
        return self._com_object.ImageArray

    @property
    def ImageArrayVariant(self):
        logger.debug("_MaximCameraImageArrayVariant called")
        return float(self._com_object.ImageArray)

    @property
    def ImageReady(self):
        logger.debug("_MaximCameraImageReady called")
        return self._com_object.ImageReady

    @property
    def IsPulseGuiding(self):
        logger.debug("_MaximCameraIsPulseGuiding called")
        raise NotImplementedError

    @property
    def LastExposureDuration(self):
        logger.debug("_MaximCameraLastExposureDuration called")
        return self._last_exposure_duration

    @property
    def LastExposureStartTime(self):
        logger.debug("_MaximCameraLastExposureStartTime called")
        return self._last_exposure_start_time

    @property
    def MaxADU(self):
        logger.debug("_MaximCameraMaxADU called")
        raise NotImplementedError

    @property
    def MaxBinX(self):
        logger.debug("_MaximCameraMaxBinX called")
        return self._com_object.MaxBinX

    @property
    def MaxBinY(self):
        logger.debug("_MaximCameraMaxBinY called")
        return self._com_object.MaxBinY

    @property
    def NumX(self):
        logger.debug("_MaximCameraNumX called")
        return self._com_object.NumX

    @NumX.setter
    def NumX(self, value):
        logger.debug(f"NumX setter called with value={value}")
        self._com_object.NumX = value

    @property
    def NumY(self):
        logger.debug("_MaximCameraNumY called")
        return self._com_object.NumY

    @NumY.setter
    def NumY(self, value):
        logger.debug(f"NumY setter called with value={value}")
        self._com_object.NumY = value

    @property
    def Offset(self):
        logger.debug("_MaximCameraOffset called")
        raise NotImplementedError

    @Offset.setter
    def Offset(self, value):
        logger.debug(f"Offset setter called with value={value}")
        raise NotImplementedError

    @property
    def OffsetMax(self):
        logger.debug("_MaximCameraOffsetMax called")
        raise NotImplementedError

    @property
    def OffsetMin(self):
        logger.debug("_MaximCameraOffsetMin called")
        raise NotImplementedError

    @property
    def Offsets(self):
        logger.debug("_MaximCameraOffsets called")
        raise NotImplementedError

    @property
    def PercentCompleted(self):
        logger.debug("_MaximCameraPercentCompleted called")
        raise NotImplementedError

    @property
    def PixelSizeX(self):
        logger.debug("_MaximCameraPixelSizeX called")
        return self._com_object.PixelSizeX

    @property
    def PixelSizeY(self):
        logger.debug("_MaximCameraPixelSizeY called")
        return self._com_object.PixelSizeY

    @property
    def ReadoutMode(self):
        logger.debug("_MaximCameraReadoutMode called")
        return self._com_object.ReadoutMode

    @ReadoutMode.setter
    def ReadoutMode(self, value):
        logger.debug(f"ReadoutMode setter called with value={value}")
        self._com_object.ReadoutMode = value

    @property
    def ReadoutModes(self):
        logger.debug("_MaximCameraReadoutModes called")
        return self._com_object.ReadoutModes

    @property
    def SensorName(self):
        logger.debug("_MaximCameraSensorName called")
        raise NotImplementedError

    @property
    def SensorType(self):
        logger.debug("_MaximCameraSensorType called")
        raise NotImplementedError

    @property
    def SetCCDTemperature(self):
        logger.debug("_MaximCameraSetCCDTemperature called")
        return self._com_object.TemperatureSetpoint

    @SetCCDTemperature.setter
    def SetCCDTemperature(self, value):
        logger.debug(f"SetCCDTemperature setter called with value={value}")
        self._com_object.TemperatureSetpoint = value

    @property
    def StartX(self):
        logger.debug("_MaximCameraStartX called")
        return self._com_object.StartX

    @StartX.setter
    def StartX(self, value):
        logger.debug(f"StartX setter called with value={value}")
        self._com_object.StartX = value

    @property
    def StartY(self):
        logger.debug("_MaximCameraStartY called")
        return self._com_object.StartY

    @StartY.setter
    def StartY(self, value):
        logger.debug(f"StartY setter called with value={value}")
        self._com_object.StartY = value

    @property
    def SubExposureDuration(self):
        logger.debug("_MaximCameraSubExposureDuration called")
        raise NotImplementedError

    @SubExposureDuration.setter
    def SubExposureDuration(self, value):
        logger.debug(f"SubExposureDuration setter called with value={value}")
        raise NotImplementedError


class _MaximFilterWheel(FilterWheel):
    def __init__(self, maxim_camera):
        logger.debug("_MaximFilterWheelMaximFilterWheel __init__ called")
        self.maxim_camera = maxim_camera

    @property
    def Connected(self):
        logger.debug("_MaximFilterWheelConnected called")
        return self.maxim_camera.Connected and self.maxim_camera.HasFilterWheel

    @Connected.setter
    def Connected(self, value):
        logger.debug(f"Connected setter called with value={value}")
        self.maxim_camera.Connected = value

    @property
    def Name(self):
        logger.debug("_MaximFilterWheelName called")
        self.maxim_camera.FilterWheelName

    @property
    def FocusOffsets(self):
        logger.debug("_MaximFilterWheelFocusOffsets called")
        raise NotImplementedError

    @property
    def Names(self):
        logger.debug("_MaximFilterWheelNames called")
        return self.maxim_camera.FilterNames

    @property
    def Position(self):
        logger.debug("_MaximFilterWheelPosition called")
        return self.maxim_camera.Filter

    @Position.setter
    def Position(self, value):
        logger.debug(f"Position setter called with value={value}")
        self.maxim_camera.Filter = value


class _MaximPinpointWCS(WCS):
    def __init__(self, maxim_camera):
        logger.debug("_MaximPinpointWCS __init__ called")
        self.maxim_camera = maxim_camera

    def Solve(self, ra=None, dec=None, scale_est=None, *args, **kwargs):
        logger.debug(
            f"_MaximPinpointWCS.Solve called with ra={ra}, dec={dec}, and scale_est={scale_est}"
        )
        self.maxim_camera.document.PinpointSolve(ra, dec, scale_est, scale_est)

        while self.maxim_camera.document.PinpointStatus == 3:
            time.sleep(1)

        if self.maxim_camera.document.PinpointStatus == 2:
            return True
        else:
            return False



from abc import ABC, abstractmethod

from ._docstring_inheritee import _DocstringInheritee


class Focuser(ABC, metaclass=_DocstringInheritee):
    @abstractmethod
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    def Halt(self):
        pass

    @abstractmethod
    def Move(self, Position):
        pass

    @property
    @abstractmethod
    def Absolute(self):
        pass

    @Absolute.setter
    @abstractmethod
    def Absolute(self, value):
        pass

    @property
    @abstractmethod
    def IsMoving(self):
        pass

    @property
    @abstractmethod
    def MaxIncrement(self):
        pass

    @property
    @abstractmethod
    def MaxStep(self):
        pass

    @property
    @abstractmethod
    def Position(self):
        pass

    @property
    @abstractmethod
    def StepSize(self):
        pass

    @property
    @abstractmethod
    def TempComp(self):
        pass

    @TempComp.setter
    @abstractmethod
    def TempComp(self, value):
        pass

    @property
    @abstractmethod
    def TempCompAvailable(self):
        pass

    @property
    @abstractmethod
    def Temperature(self):
        pass

    @Temperature.setter
    @abstractmethod
    def Temperature(self, value):
        pass



import numpy as np

from pyscope.utils import airmass


def test_airmass():
    deg = np.pi / 180
    assert np.round(airmass(90 * deg), 2) == 1.0
    assert np.round(airmass(60 * deg), 2) == 1.15
    assert np.round(airmass(45 * deg), 2) == 1.41
    assert np.round(airmass(30 * deg), 2) == 1.99
    assert np.round(airmass(15 * deg), 2) == 3.81
    assert np.round(airmass(0 * deg), 2) == 38.75
